# Generic Makefile for a single executable program that uses the PBM libraries
# Last edited on 2007-11-01 12:57:38 by stolfi

######################################################################
# To be included by the top-level Makefile in the program's directory.
# The caller must define ${PROG} or ${PROGS}, and possibly also
# ${BAD_PROGS}, ${JSLIBS}, ${PBMLIBS}

ifeq "/${PROGS}" "/"
PROGS := ${PROG}
endif

ifneq "/${PROGS}" "/"

SHELL := /bin/sh

# Where to install the program
INSTPUB := ${HOME}/PUB
INSTINC := ${INSTPUB}/include
INSTLIB := ${INSTPUB}/${PLATFORM}/lib
INSTBIN := ${INSTPUB}/${PLATFORM}/bin
INSTMAN := ${INSTPUB}/${PLATFORM}/man
INSTCAT := ${INSTPUB}/${PLATFORM}/cat

# Location of Stolfi's libraries
JSPUB := ${HOME}/PUB
JSLIB := ${JSPUB}/${PLATFORM}/lib
JSINC := ${JSPUB}/include

# Get list of sources:
HFILES := ${shell shopt -s nullglob; ls *.h Makefile | egrep -v '^Makefile$$'}
CFILES := ${shell shopt -s nullglob; ls *.c Makefile | egrep -v '^Makefile$$'}

# Get list of main and non-main sources:
PROG_CFILES := ${addsuffix .c,${PROGS}}
BAD_PROG_CFILES := ${addsuffix .c,${BAD_PROGS}}
LIB_CFILES := ${filter-out ${PROG_CFILES} ${BAD_PROG_CFILES}, ${CFILES}}

# Derived objects:
HOFILES := ${subst .h,.ho,${HFILES}}
PROG_OFILES := ${subst .c,.o,${PROG_CFILES}}
BAD_PROG_OFILES := ${subst .c,.o,${BAD_PROG_CFILES}}
LIB_OFILES := ${subst .c,.o,${LIB_CFILES}}

# Make targets:
.PHONY: \
  all depend \
  uninstall uninstall-man \
  build build-prog check \
  install install-man \
  clean

# all: uninstall clean build check install
all: uninstall build install

# File with dependencies between sources:
DEPFILE := Deps.make

# Directories to search for #include files:
IFLAGS := -I. -I${JSINC}

#---------------------------------------------------------------------
# "make depend" recreates the source dependency file ${DEPFILE}.

depend: 
	/bin/rm -f ${DEPFILE}
	extract-ho-deps ${IFLAGS} -I/usr/include \
            ${HFILES} ${CFILES} \
          | egrep -v ': /usr/include' \
          > ${DEPFILE}

#---------------------------------------------------------------------
# "make build" assumes that the dependencies in ${DEPFILE}
# are up-to-date.

build: build-prog
  
build-prog: ${PROGS}
  
LIBRARIES := \
  ${addprefix ${JSLIB}/, ${JSLIBS}} \
  ${OTHERLIBS}

CC := /usr/bin/gcc

CFLAGS := \
  -ggdb \
  -ansi \
  -Wall -Wpointer-arith -Wmissing-prototypes \
  -fpcc-struct-return \
  -ffloat-store -frounding-math
  
LDFLAGS := \
  -ggdb

%.o: %.c ;\
  ${CC} -c ${CFLAGS} ${EXTRACFLAGS} ${IFLAGS} $*.c
  
%.ho: %.h ;\
  ${CC} -o $*.ho -c ${CFLAGS} ${IFLAGS} -x c $*.h \
  || /bin/rm -f $*.ho
  
${PROGS}: ${PROG_OFILES} ${LIB_OFILES} ${LIBRARIES}
	@echo '= = = = = = = = = = = = = = = = = = = = = = = = = = = = = ='
	@echo building $@ ...
	-rm -f $@
	${CC} -o $@ ${LDFLAGS} $@.o \
          ${LIB_OFILES} \
          ${LIBRARIES} -lm -lrt
	@echo '= = = = = = = = = = = = = = = = = = = = = = = = = = = = = ='
        
# Include specific dependencies extracted by "make depend"
include ${DEPFILE}

#---------------------------------------------------------------------
# "make install" copies program and makefile to the public dir.

MANSECT := 1
MANPAGES =  ${addsuffix .${MANSECT},${PROGS}}

install: ${PROGS} install-man
	cp -p ${PROGS} ${INSTBIN}/

install-man:
	@for mp in ${MANPAGES} ; do \
          if [ -r $${mp} ]; then \
            cp -p $${mp} ${INSTMAN}/man${MANSECT}/ ;\
          else \
            echo "file $${mp} not found - ignored" ;\
          fi ; \
        done

#---------------------------------------------------------------------
# "make uninstall" deletes the exported program and makefile:

uninstall: uninstall-man
	( cd ${INSTBIN} && rm -f ${PROGS} )

uninstall-man:
	( cd ${INSTMAN}/man${MANSECT} && rm -f ${MANPAGES} )
	( cd ${INSTCAT}/cat${MANSECT} && rm -f ${MANPAGES} )  

#---------------------------------------------------------------------
# "make archive" creates a tarball of sources, manpages, and executables.

NOW = ${shell date '+%Y-%m-%d-%H%M%S'}
TARBALL_NAME := ${NOW}.tgz
  
archive: ;\
  tar -cvzf ${TARBALL_NAME} \
          00-README Makefile \
            ${PROGS} \
            ${HFILES} ${CFILES} \
            ${MANPAGES}

#---------------------------------------------------------------------
# "make check" runs the tests in the "tests" subdirectory.

TESTDIR := tests

check: ${PROGS}
	( cd ${TESTDIR} && ${MAKE} all )

#---------------------------------------------------------------------
# "make clean" deletes all derived files.

clean:
	-/bin/rm -f *.o *.ho *.a core ${PROGS} 
	( cd ${TESTDIR} && ${MAKE} clean )

endif
# End of ${PROGS} section.
######################################################################

