/* See dm_cand.h */
/* Last edited on 2007-01-04 01:57:12 by stolfi */

#define dm_cand_C_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)"

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_rung.h>
#include <dm_pairing.h>
#include <dm_cand.h>
#include <dm_cand_list.h>
#include <dm_image.h>

#include <affirm.h>
#include <fget.h>
#include <filefmt.h>
#include <jsrandom.h>
#include <jsmath.h>

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

/* INTERNAL PROTOTYPES */

void dm_cand_write_painting_as_pgm(float_image_t *img, char *name, char *tag);
  /* Writes the image {img} to disk as a PGM file called "{name}{tag}.pgm".
    The entries of {img} are scaled so that the smallest pixel maps to gray,
    the largest pixel maps to black.  IN particular, pixels with value
    {-INF} map to white. */

/* EXPORTED IMPLEMENTATIONS */

vec_typeimpl(dm_cand_vec_t,dm_cand_vec,dm_cand_t);

bool_t dm_cand_is_valid(dm_cand_t *cd)
  {
    /* Get and check the pairing: */
    dm_pairing_t *p = cd->pr;
    (void)dm_pairing_is_valid(p, TRUE);
    if (! dm_pairing_is_unitary_monotonic(p)) { return FALSE; }
    int ng = dm_pairing_num_rungs(p);
    if (ng != 0) 
      { /* Get datum counts of the two sequences: */
        int nx = cd->nbas[0];
        int ny = cd->nbas[1];

        /* The sequences must be non-empty: */
        if (nx <= 0) { return FALSE; }
        if (ny <= 0) { return FALSE; }

        /* The initial rung must lie in the initial cell: */
        dm_rung_t gini = dm_pairing_get_rung(p, 0);
        int ix = gini.c[0], iy = gini.c[1];
        if ((ix < 0) || (ix >= nx)) { return FALSE; }
        if ((iy < 0) || (iy >= ny)) { return FALSE; }

        /* The final rung must be greater than or equal to the first one: */
        dm_rung_t gfin = dm_pairing_get_rung(p, ng-1);
        int fx = gfin.c[0], fy = gfin.c[1];
        if (fx < ix) { return FALSE; }
        if (fy < iy) { return FALSE; }

        if (dm_pairing_is_circular(p))
          { /* Circular pairing. */
            /* Both sequences must be circular: */
            if ((! cd->circ[0]) || (! cd->circ[1]))  { return FALSE; }
            /* Get the coordinate period vector {(DX,DY)}: */
            dm_rung_t rper = dm_pairing_period(p);
            int DX = rper.c[0], DY = rper.c[1];
            /* X- or Y-stationary cands are not allowed. */
            if ((DX == 0) || (DY == 0)) { return FALSE; }
            /* Each coordinate period must be one or more sweeps over the sequence: */
            if ((DX <= 0) || (DX % nx != 0)) { return FALSE; }
            if ((DY <= 0) || (DY % ny != 0)) { return FALSE; }
          }
        else
          { /* Open pairing. */
            /* If either sequence is open, the pairing must not fall off its end: */
            if ((! cd->circ[0]) && (fx >= nx)) { return FALSE; }
            if ((! cd->circ[1]) && (fy >= ny)) { return FALSE; }
          }
      }
    return TRUE;
  }      

dm_cand_t dm_cand_from_pairing
  ( dm_seq_t *xp, 
    dm_seq_t *yp, 
    dm_pairing_t *pr,
    double score
  )
  { /* Get the datum counts {nx,ny} in the two sequences: */
    int nx = xp->dv.nel; demand(nx > 0, "bad {nx}");
    int ny = yp->dv.nel; demand(ny > 0, "bad {ny}");
    /* Fill fields {id,name,nbas,circ,pr,score} of candidate: */
    dm_cand_t cd;
    cd.id[0] = xp->id; cd.name[0] = xp->name; cd.nbas[0] = nx; cd.circ[0] = xp->circ;
    cd.id[1] = yp->id; cd.name[1] = yp->name; cd.nbas[1] = ny; cd.circ[1] = yp->circ;
    cd.pr = pr;
    cd.score = score;
    return cd;
  }

typedef double dm_subpairing_score_proc_t(int64_t r, int64_t s, bool_t circsub); 
  /* A client procedure that computes a score for a subpairing {p[r..s]}
    of some perfect pairing {p}.  If {circsub} the sub-pairing is circular
    and equal to the whole of {p},  which should be circular too. */

typedef void dm_subpairing_use_proc_t
  ( int64_t r,
    int64_t s,
    bool_t circsub,
    double score
  ); 
  /* A client procedure that uses a subpairing {p[r..s]} of some
    perfect pairing {p}, whose score is {score}. If {circsub} the
    subpairing is circular and equal to the whole of {p}, which should
    be circular too. */

void dm_cand_scan_alignment
  ( int64_t ng,         /* Number of pairs in alignment. */
    bool_t circp,        /* TRUE if alignment is circular. */ 
    dm_subpairing_score_proc_t *score,
    int64_t minRungs,   /* Min number of rungs in a valid subpairing. */
    int64_t maxRungs,   /* Max number of rungs in a valid subpairing. */
    dm_subpairing_use_proc_t *use  /* What to do with candidates. */
  );
  /* Scans the perfect pairing {p} of two sequences {x,y} that has
    {ng} rungs, and adds all candidates found in it to the heap. If
    {circp} the pairing {p} is assumed to be circular.

    A valid sub-pairing is a set of consecutive rungs from {p}
    that has at least {minRungs} and at most {maxRungs} rungs. A
    candidate is a valid sub-pairing whose score is positive and
    better than that of any of proper valid sub-pairing or
    super-pairing. 
    
    The score of a non-circular subpairing {p[r..s]} is assumed to be
    given by {score(r,s,FALSE)}. If {p} is circular, the score of {p}
    should be given by {score(0,ng-1,TRUE)}. */

#define dm_cand_checking_level 1
  /* Define this as positive to get paranoid checking. */

#define CHKLEV(n) (dm_cand_checking_level >= (n))
  /* Use {if(CHKLEV(n)){...}} for checks of paranoia level {n}. */

#define dm_cand_debug_level 2
  /* Set this positive to get some debugging printouts. */

#define DEBLEV(n) (dm_cand_debug_level >= (n))
  /* Use {if(DEBLEV(n)){...}} for debug printouts of level {n}. */

int dm_cand_count_common_rungs(dm_cand_t *ca, dm_cand_t *cb)
  { if (ca->id[0] != cb->id[0]) { return 0; }
    if (ca->id[1] != cb->id[1]) { return 0; }
    demand(ca->nbas[0] == cb->nbas[0], "inconsistent seq lengths"); 
    demand(ca->nbas[1] == cb->nbas[1], "inconsistent seq lengths");
    int nx = ca->nbas[0], ny = ca->nbas[1];
    return dm_pairing_count_common_rungs(ca->pr, cb->pr, nx, ny);
  }
  
dm_cand_vec_t dm_cand_get_best_perfect
  ( dm_seq_t *xp,
    dm_seq_t *yp,
    dm_step_score_proc_t *step_score,
    int minBases, 
    double minScore,
    int maxCands,
    double frac
  )
  {
    int nx = xp->dv.nel; bool_t circx = xp->circ;
    int ny = yp->dv.nel; bool_t circy = yp->circ;
    
    int64_t minRungs = minBases; 
      /* Min number of rungs in a candidate, including the circular one. */
    int64_t maxRungs = (nx < ny ? nx : ny);  
      /* Max number of rungs in a candidate, including the circular one. */
    
    /* Working vectors for {dm_cand_scan_alignment}: */
    double_vec_t rsc = double_vec_new(0);
    double_vec_t psc = double_vec_new(0);

    dm_cand_vec_t cdv = dm_cand_vec_new(maxCands);
    int ncand = 0;
    /* The current set of candidates is kept in the array
      {cdv[0..ncand-1]}. During the scan, the array is sorted
      by *decreasing* score (i.e. the worst candidate is 
      {cdv[ncand-1]}. */

    auto void scan_alignment(int ix, int iy, int64_t ng, bool_t circp);
      /* Scans the perfect pairing {p} that starts with rung {(ix,iy)}
        and has {ng} rungs, and adds all candidates found in it to the
        heap. If {circp} the pairing is assumed to be circular.
        
        A valid sub-pairing is a set of consecutive rungs from {p}
        that has at least {minBases} and at most {min(nx,ny)} rungs. A
        candidate is a valid sub-pairing whose score is positive and
        better than that of any of proper valid sub-pairing or
        super-pairing. */

    /* Scan all maximal-length perfect pairings, collect best cands: */
    dm_pairing_enum_alignments(nx, circx, ny, circy, &scan_alignment); 
    
    /* Release working storage: */
    double_vec_trim(&rsc, 0);
    double_vec_trim(&psc, 0);

    /* Trim the candidate vector to the number of cands actually found: */
    dm_cand_vec_trim(&cdv, ncand);
    
    return cdv; 

    void scan_alignment(int ix, int iy, int64_t ng, bool_t circp)
      {
        demand(ng > 0, "bad {ng}");

        auto double subpairing_score(int64_t r, int64_t s, bool_t circsub);
          /* Score of sub-pairing {p[r..s]}, namely
            {S = SUM{step_score(p[j-1],p[j]) : j \in r+1..s}}.
            If {circsub} adds also {step_score(p[s],p[r])},
            else adds also none-to-first-rung and last-rung-to-none. 
            Obtained from precomputed tables.
          */

        auto double pref_score(int64_t k);
          /* Score of non-circular sub-pairing {p[0..k]}, from tables. */

        auto double join_score(int64_t k);
          /* Score of rung {p[k]}, from tables. */

        auto void initialize_score_tables(void);  
          /* Initializes the tables used by {score}. */
          
        auto void check_all_scores(void); 
          /* Checks {subpairing_score(r,s)} against defintion for all valid {(r,s)}. */
         
        auto void check_score(int64_t r, int64_t s);  
          /* Checks {subpairing_score(r,s)} against defintion. */

        auto void store_candidate(int64_t r, int64_t s, bool_t circsub, double score);
          /* Stores into the sorted list {cdv} the pairing
            {(ix+k,iy+k)} for {k=r,..s}, where both indices are taken
            modulo the respective sequence lengths. If {circsub},
            assumes that the pairing is circular. Assumes that its
            score is {score}. The procedure is a no-op if the pairing
            is not among the best {maxCands} entries found so far, or
            is already in the list. */
        
        if (DEBLEV(1))
          { fprintf(stderr, "\nalignment (%d,%d:%lld)\n", ix, iy, ng); }
            
        /* Allocate score tables: */
        initialize_score_tables();

        /* Paranoia checks for the {subpairing_score} function: */
        if (CHKLEV(2)) { check_all_scores(); }

        dm_cand_scan_alignment
          ( ng, circp, 
            &subpairing_score,
            minRungs, 
            maxRungs, 
            &store_candidate
          );
          
        auto void update_score_tables(int64_t kcur);
          /* Update the table of precomputed scores so that it is valid for {k} 
            up to {kcur}.  May discard some older entries. */

        int64_t kmin, kmax, ksize;
        /* The tables {rsc,psc} are used to store precomputed rung and
          total scores for this alignment. The score of rung {p[k]} is
          stored in {rsc[k % ksize]}. The score of the (non-circular)
          sub-pairing {p[0..k]} is stored in {psc[k % ksize]}. These data
          are valid only for {k} in {kmin..kmax}. */

        void initialize_score_tables(void)
          { ksize = 2*maxRungs - minRungs + 1;
            double_vec_expand(&rsc, ksize);
            double_vec_expand(&psc, ksize);
            dm_rung_t g = (dm_rung_t){{ix, iy}};
            rsc.el[0] = step_score(dm_rung_none, g, xp, yp);
            psc.el[0] = rsc.el[0];
            kmin = 0; kmax = 0; 
          }

        void update_score_tables(int64_t kcur)
          { 
            demand(kcur >= kmin, "table is too short");
            if (kcur <= kmax) { return; }
            dm_rung_t g = (dm_rung_t){{ix + kmax, iy + kmax}};
            do
              { int64_t k = kmax + 1;
                dm_rung_t h = (dm_rung_t){{ix + k, iy + k}};
                double ssck = (step_score == NULL ? 0 : step_score(g, h, xp, yp));
                int kk = k % ksize;
                psc.el[kk] = psc.el[kmax % ksize] + ssck;
                g = h;
                kmax++;
                if ((kmax - kmin) > ksize) { kmin++; }
              }
            while (kmax < kcur);
          }

        double pref_score(int64_t k)
          { demand(k >= 0, "invalid prefix: k < 0");
            demand(circp || (k < ng), "invalid prefix: out of range");
            /* Table consistency: */
            update_score_tables(k);
            affirm((kmin <= k) && (k <= kmax), "out of table range");
            return psc.el[k % ksize]; 
          }

        double join_score(int64_t k)
          { demand(k >= 0, "invalid prefix: k < 0");
            demand(circp || (k < ng), "invalid prefix: out of range");
            /* Table consistency: */
            update_score_tables(k);
            affirm((kmin <= k) && (k <= kmax), "out of table range");
            return rsc.el[k % ksize]; 
          }

        double subpairing_score(int64_t a, int64_t b, bool_t circsub)
          { /* Argument consistency: */
            /* fprintf(stderr, "  sc a = %6lld  b = %6lld %c", a, b, (circsub ? 'C' : 'N')); */
            demand(a <= b, "invalid sub-pairing: a > b");
            demand(a >= 0, "invalid sub-pairing: a < 0");
            demand(circp || (b < ng), "invalid sub-pairing: out of range");
            demand(b - a + 1 <= maxRungs, "invalid sub-pairing: too long");
            double s;
            if (circsub)
              { assert((b - a) == ng - 1);
                s = pref_score(b + 1) - pref_score(a);
              }
            else
              { s = pref_score(b) - pref_score(a) + join_score(a); }
            /* fprintf(stderr, " = %24.16e\n", s); */
            return s;
          }
          
        void check_score(int64_t r, int64_t s)
          { double eps = 1.0e-10; /* Absolute error allowed. */
            double tol = 1.0e-6;  /* Relative error allowed. */
            double s_slow = dm_pairing_perfect_score(xp, ix+r, yp, iy+r, s-r+1, FALSE, step_score);
            double s_fast = subpairing_score(r,s, FALSE);
            double s_size = fabs(s_slow) + fabs(s_fast) + eps;
            if (fabs(s_slow - s_fast) > eps + tol*s_size)
              {
                fprintf(stderr, "bad {score}\n");
                fprintf(stderr, "  fast = %24.26e\n", s_fast);
                fprintf(stderr, "  slow = %24.26e\n", s_slow);
                assert(FALSE);
              }
          }

        void check_all_scores(void)
          { fprintf(stderr, "checking the {subpairing_score} function...\n");
            int64_t rt, st;
            for (rt = 0; rt < ng; rt++)
              { 
                int64_t stmax = rt + maxRungs - 1;
                if ((! circp) && (stmax >= ng)) { stmax = ng - 1; }
                for (st = rt + minRungs - 1; st <= stmax; st++)
                  {  check_score(rt, st); }
                fprintf(stderr, "#"); 
              }
            fprintf(stderr, "\n"); 
          }

        void store_candidate(int64_t r, int64_t s, bool_t circsub, double score)
          { /* Reduce the initial indices modulo the respective periods: */
            dm_rung_t gini = (dm_rung_t){{ (ix + r) % nx, (iy + r) % ny }};
            int64_t ns = s - r + 1;
            if (DEBLEV(2))
              { fprintf
                  ( stderr, 
                    "\n  perfect candidate starting at (%d,%d) steps = %lld score = %24.16e\n", 
                    gini.c[0], gini.c[1], ns, score
                  ); 
              }
            /* Worth adding: */
            if (score < minScore) { return; }
            /* Compute the coordinate period vector: */
            if (circsub) { demand(ns == lcm(nx,ny), "invalid perfect circular pairing"); }
            dm_pairing_t *pr = dm_pairing_perfect(&gini, ns, circsub);
            dm_cand_t cd = dm_cand_from_pairing(xp, yp, pr, score);
            dm_cand_list_insert(&cdv, &ncand, maxCands, &cd, frac);
          }
      }
  }

void dm_cand_scan_alignment
  ( int64_t ng,         /* Number of pairs in alignment. */
    bool_t circp,       /* TRUE if alignment is circular. */ 
    dm_subpairing_score_proc_t *score,
    int64_t minRungs,   /* Min number of rungs in a valid subpairing. */
    int64_t maxRungs,   /* Max number of rungs in a valid subpairing. */
    dm_subpairing_use_proc_t *use  /* What to do with candidates. */
  )
  {
    /* Trivial cases: */
    if ((ng < minRungs) || (maxRungs < minRungs)) { return; }
    
    auto void initialize_mxsma(int64_t ri, int64_t si);
      /* Initialize the {mxsma} table for the current subpairing {[ri,si]}: */
    
    auto void initialize_mxbig(int64_t ri, int64_t si);
      /* Initialize the {mxbig} table for the current subpairing {[ri,si]}: */

    auto double *mxsma(int64_t r, int64_t s);
      /* The maximum score among all valid sub-pairings {[a,b]} that 
        are contained in or equal to {[r,s]}. */
      
    auto double *mxbig(int64_t r, int64_t s);
      /* The maximum score among all valid  subpairings {[a,b]} 
        that properly contain {[r,s]}. */
    
    auto void update_mxsma(int64_t rn, int64_t sn);
      /* Recompute {mxsma(t,sn)} from {mxsma(t,sn-1)} 
        for {t} in {rn .. sn-minRungs+1}. */

    auto void update_mxbig(int64_t rn, int64_t sn);
      /* Recompute {mxbig(rn,t)} from {mxbig(rn-1,t)} 
        for {t} in {sn .. rn + maxRungs - 1}. */

    auto void check_mxsma(int64_t rn, int64_t sn);
      /* Checks {mxsma(t,sn)} for {t} in {rn .. sn-minRungs+1}. */

    auto void check_mxbig(int64_t rn, int64_t sn);
      /* Checks {mxbig(rn,t)} for {t} in {sn .. rn + maxRungs - 1}. */

    int N = maxRungs - minRungs + 1;
    double tsma[N]; /* The value of {mxsma(r,s)} is stored in {tsma[r % N]} */
    double tbig[N]; /* The value of {mxbig(r,s)} is stored in {tbig[s % N]} */

    /* If the pairing {p} is circular, we must consider it too as a
      potential candidate. In that case, note that {p} properly
      contains any other candidate; so, if its score is better than
      that of any open sub-pairing, {p} is the only candidate. In all
      other cases {p} is not a candidate. */
      
    bool_t circ_valid = circp && (ng <= maxRungs); 
      /* TRUE if the full circular pairing is valid. */
    
    double circ_score = (circ_valid ? score(0, ng, TRUE) : -INF);
      /* Score of full circular pairing, if it is valid, else {-oo}. */
    
    bool_t circ_best = circ_valid;
      /* TRUE while the full curcular alignment is the best one. */

    /* Find the initial pre-candidate {[r,s]}.
      
      Pre-candidate condition: A valid sub-pairing {[r,s]} is a
      pre-candidate if {s} is the smallest index such that {[r,s]} is
      valid and {mxsma(r,s)> mxbig(r,s)}. A pre-candidate is a
      candidate if {f(r,s)==mxsma(r,s)}.  */
    
    int64_t rc, sc; /* Current pre-candidate is {[rc,sc]}. */
    
    /* Choose {rc} of initial pre-candidate: */
    if (! circp)
      { /* Must start at {rc = 0}. Note that {mxbig} will work fine. */
        rc = 0;
      }
    else
      { /* Start at a large enough {sc} so that {mxbig(rc,sc)} 
          can be computed without going into negative indices. */
        rc = maxRungs - minRungs;
      }
      
    /* Intialize {sc} at minimum value for this {rc}: */
    sc = rc + minRungs - 1;
      
    /* Initialize the {mxsma,mxbig} tables for this subpairing: */
    initialize_mxsma(rc,sc);
    initialize_mxbig(rc,sc);
    
    /* Scan pre-candidates and output candidates: */
    int64_t ncol;
    for (ncol = 0; ncol < ng - 1; ncol++)
      { 
        /* fprintf(stderr, "\nlooking for candidates with r = %lld\n", rc); */
        /* fprintf(stderr, "  starting with [%lld,%lld]\n", rc, sc); */
        /* Get max {sc} for this {rc}: */
        int64_t smax = rc + maxRungs - 1;
        if ((! circp) && (smax >= ng)) { smax = ng - 1; }
        assert(sc <= smax);
        /* Increment {sc} trying to satisfy the pre-candidate condition: */
        double vsma = (*(mxsma(rc,sc)));
        double vbig = (*(mxbig(rc,sc)));
        /* fprintf(stderr, "  mxsma = %24.16e  mxbig = %24.16e\n", vsma, vbig); */
        while ((sc < smax) && (vsma <= vbig)) 
          { sc++;
            /* fprintf(stderr, "  incrementing sc to [%lld,%lld]\n", rc, sc); */
            update_mxsma(rc,sc);
            vsma = (*(mxsma(rc,sc)));
            vbig = (*(mxbig(rc,sc)));
            /* fprintf(stderr, "  mxsma = %24.16e  mxbig = %24.16e\n", vsma, vbig); */
          }
        /* If {rc,sc} is a candidate, output it: */
        if (vsma > vbig)
          { double rssc = score(rc,sc,FALSE);
            if ((rssc > circ_score) && (rssc == vsma)) 
              { use(rc, sc, FALSE, rssc);
                circ_best = FALSE; /* The full circular alignment is not the best one. */
              }
          }
        /* Advance {rc}, incrementing {sc} if necessary to keep {rc,sc} valid: */
        int64_t ns = sc - rc + 1;
        if (ns <= minRungs)
          { assert(ns == minRungs);
            rc++; 
            sc++; 
            if ((! circp) && (sc >= ng)) { /* No more cands. */  break; }
            /* fprintf(stderr, "  incrementing rc and sc to [%lld,%lld]\n", rc, sc); */
            update_mxsma(rc,sc);
            update_mxbig(rc,sc);
          }
        else
          { rc++;
            /* fprintf(stderr, "  incrementing rc to [%lld,%lld]\n", rc, sc); */
            update_mxbig(rc,sc);
          }
      }
    
    /* If full circular subpairing is still the best, add it: */
    if (circ_best) { use(0, ng-1, TRUE, circ_score); }

    void initialize_mxsma(int64_t ri, int64_t si)
      { int64_t rmax = si - minRungs + 1;
        assert(rmax >= ri);
        /* Compute {mxsma(t,si)} for all {t} ii {ri..rmax}: */
        /* fprintf(stderr, "  initializing mxsma(%lld..%lld,%lld)\n", ri, rmax, si); */
        int64_t smin = ri + minRungs - 1;
        int64_t s1;
        for (s1 = smin; s1 <= si; s1++) { update_mxsma(ri, s1); }
      }
          
    void initialize_mxbig(int64_t ri, int64_t si)
      {
        int64_t smax = ri + maxRungs - 1;
        if ((! circp) && (smax >= ng)) { smax = ng; }
        assert(smax >= si);
        /* Compute {mxbig(ri,t)} for {t} ii {si..smax}: */
        /* fprintf(stderr, "  initializing mxbig(%lld,%lld..%lld)\n", ri, si, smax); */
        int rmin = si - maxRungs + 1;
        if ((! circp) && (rmin < 0)) { rmin = 0; }
        assert(rmin >= 0); /* If {circ} caller must use {ri} large enough to avoid negs. */
        int64_t r1;
        for (r1 = rmin; r1 <= ri; r1++) { update_mxbig(r1, si); }
      }

    double *mxsma(int64_t r, int64_t s)
      { assert(r >= 0); return &(tsma[r % N]); }
      
    double *mxbig(int64_t r, int64_t s)
      { assert(s >= 0); return &(tbig[s % N]); }
      
    void update_mxsma(int64_t rn, int64_t sn)
      { assert(rn >= 0);
        assert(sn >= 0);
        double scmax = -INF;
        int64_t rmax = sn - minRungs + 1; 
        assert(rmax >= rn);
        int64_t rt;
        for (rt = rmax; rt >= rn; rt--)
          { 
            double s;
            s = score(rt, sn, FALSE);
            if (s > scmax) { scmax = s; }
            if (rt < rmax)
              { s = (*(mxsma(rt,sn-1)));
                if (s > scmax) { scmax = s; }
              }
            /* fprintf(stderr, "  updating mxsma(%lld,%lld) = %24.16e\n", rt, sn, scmax); */
            (*(mxsma(rt,sn))) = scmax;
          }
        if (CHKLEV(1)) { check_mxsma(rn, sn); }
      }
      
    void check_mxsma(int64_t rn, int64_t sn)
      { assert(rn >= 0);
        assert(sn >= 0);
        double scmax = -INF;
        int64_t rmax = sn - minRungs + 1; 
        assert(rmax >= rn);
        /* fprintf(stderr, "  checking mxsma(%lld..%lld,%lld)\n", rn, rmax, sn); */
        int64_t rt;
        for (rt = rmax; rt >= rn; rt--)
          { int64_t smin = rt + minRungs - 1;
            assert(smin >= 0);
            assert(smin <= sn);
            int64_t st;
            for (st = smin; st <= sn; st++)
              { double s = score(rt, st, FALSE);
                if (s > scmax) { scmax = s; }
              }
            assert((*(mxsma(rt,sn))) == scmax);
          }
      }
      
    void update_mxbig(int64_t rn, int64_t sn)
      { assert(rn >= 0);
        assert(sn >= 0);
        double scmax = -INF;
        int64_t smax = rn + maxRungs - 1; 
        if ((! circp) && (smax >= ng)) { smax = ng - 1; }
        assert(smax >= sn);
        int64_t st;
        for (st = smax; st >= sn; st--)
          { 
            int64_t rmin = st - maxRungs + 1;
            if ((! circp) && (rmin < 0)) { rmin = 0; }
            assert(rmin >= 0); /* If {p} is circular, {rn} must be large enough to avoid negs. */
            double s;
            if (rn > rmin)
              { s = score(rn-1, st, FALSE);
                if (s > scmax) { scmax = s; }
                s = (*(mxbig(rn-1,st)));
                if (s > scmax) { scmax = s; }
              }
            /* fprintf(stderr, "  updating mxbig(%lld,%lld) = %24.16e\n", rn, st, scmax); */
            (*(mxbig(rn,st))) = scmax;
            s = score(rn, st, FALSE);
            if (s > scmax) { scmax = s; }
          }
        if (CHKLEV(1)) { check_mxbig(rn, sn); }
      }

    void check_mxbig(int64_t rn, int64_t sn)
      { assert(rn >= 0);
        assert(sn >= 0);
        double scmax = -INF;
        int64_t smax = rn + maxRungs - 1; 
        if ((! circp) && (smax >= ng)) { smax = ng - 1; }
        assert(smax >= sn);
        /* fprintf(stderr, "  checking mxbig(%lld,%lld..%lld)\n", rn, sn, smax); */
        int64_t st;
        for (st = smax; st >= sn; st--)
          { int64_t rmin = st - maxRungs + 1;
            if ((! circp) && (rmin < 0)) { rmin = 0; }
            assert(rmin >= 0); /* If {p} is circular, {rn} must be large enough to avoid negs. */
            int64_t rt;
            double s;
            for (rt = rmin; rt < rn; rt++)
              { s = score(rt, st, FALSE);
                if (s > scmax) { scmax = s; }
              }
            assert((*(mxbig(rn,st))) == scmax);  
            s = score(rn, st, FALSE);
            if (s > scmax) { scmax = s; }
          }
      }
  }

dm_cand_t dm_cand_map_to_finer(dm_cand_t *cdold, int nxnew, int nynew, int nwtb)
  { fprintf(stderr, "    mapping candidate\n");
    fprintf(stderr, "    old: ");
    dm_cand_write(stderr, cdold, 5, 8, 6);
    fprintf(stderr, "\n");
    int nxold = cdold->nbas[0];
    int nyold = cdold->nbas[1];
    
    dm_cand_t cdnew;
    cdnew.score = cdold->score; 
    int j;
    cdnew.nbas[0] = nxnew;
    cdnew.nbas[1] = nynew;
    for (j = 0; j <= 1; j++)
      { cdnew.id[j] = cdold->id[j];
        cdnew.name[j] = cdold->name[j];
        cdnew.circ[j] = cdold->circ[j];
      }
    /* Map pairing: */
    dm_pairing_t *prold = cdold->pr;
    dm_pairing_t *prmap = dm_pairing_map_to_finer(prold, nxold, nyold, nxnew, nynew, nwtb);
    cdnew.pr = dm_pairing_interpolate(prmap);

    fprintf(stderr, "    new: ");
    dm_cand_write(stderr, &cdnew, 5, 8, 6);
    fprintf(stderr, "\n");
    return cdnew; 
  }

dm_cand_vec_t dm_cand_vec_map_to_finer(dm_cand_vec_t *cdvold, int nxnew, int nynew, int nwtb)
  { fprintf(stderr, "  mapping %d candidates ...\n", cdvold->nel);
    dm_cand_vec_t cdvnew = dm_cand_vec_new(cdvold->nel);
    int ic;
    for (ic = 0; ic < cdvold->nel; ic++)
      { cdvnew.el[ic] = dm_cand_map_to_finer(&(cdvold->el[ic]), nxnew, nynew, nwtb); }
    fprintf(stderr, "  mapped %d candidates.\n", cdvnew.nel);
    return cdvnew;
  }

dm_cand_t dm_cand_throw
  ( dm_seq_id_t ida, 
    char *namea, 
    int na, 
    bool_t circa, 
    dm_seq_id_t idb, 
    char *nameb, 
    int nb,
    bool_t circb,
    int len,
    bool_t circp,
    bool_t diagonal,
    double skipProb
  )
  { dm_cand_t cd;
    cd.score = dgaussrand();
    cd.id[0] = ida; cd.id[1] = idb;
    cd.name[0] = namea; cd.name[1] = nameb;
    cd.nbas[0] = na; cd.nbas[1] = nb;
    cd.pr = dm_pairing_throw(na, circa, nb, circb, len, circp, diagonal, skipProb);
    return cd;
  }
  
dm_cand_vec_t dm_cand_vec_throw
  ( int ncd,
    dm_seq_id_t ida, 
    char *namea, 
    int na, 
    bool_t circa, 
    dm_seq_id_t idb, 
    char *nameb, 
    int nb, 
    bool_t circb,
    int minlen,
    int maxlen,
    double circProb,
    double diagProb,
    double skipProb
  )
  { /* Allocate the candidate vector {cdv}: */
    dm_cand_vec_t cdv = dm_cand_vec_new(ncd);
    int i; 
    for (i = 0; i < ncd; i++) 
      { /* Choose between circular and open pairing: */
        bool_t circP = circa & circb & (drandom() < circProb);
        /* Choose between near-diagonal and random: */
        bool_t diagP = (drandom() < diagProb);
        /* Choose length of candidate: */
        int len = abrandom(minlen, maxlen);
        /* Generate random candidate: */
        cdv.el[i] = dm_cand_throw
          ( ida, namea, na, circa, 
            idb, nameb, nb, circb,
            len, circP, diagP, 0.10
          );
      }
    return cdv;
  }


void dm_cand_read(FILE *rd, dm_cand_t *cd)
  {
    /* Read internal id {id}, external id {name}, and length {nbas} of each seq: */
    int j;
    for (j = 0; j <= 1; j++)
      { cd->id[j] = fget_int(rd);
        cd->name[j] = fget_string(rd);
        cd->nbas[j] = fget_int(rd);
      }
    /* Read candidate's score: */
    cd->score = fget_double(rd);
    /* Read candidate's pairing: */
    cd->pr = dm_pairing_read(rd);
  }

void dm_cand_debug(char *tag, dm_cand_t *cd)
  {
    fprintf(stderr, "  %s ", tag);
    dm_cand_write(stderr, cd, 5, 8, 6);
    fprintf(stderr, "\n");
  }

void dm_cand_write(FILE *wr, dm_cand_t *cd, int idsize, int namesize, int ixsize)
  { int j;
    for (j = 0; j <= 1; j++)
      { fprintf(wr, " %*d", idsize, cd->id[j]);
        fprintf(wr, " %-*s", namesize, cd->name[j]);
        fprintf(wr, " %*d", ixsize, cd->nbas[j]);
      }
    fprintf(wr, " %16.8f", cd->score); 
    dm_pairing_write(wr, cd->nbas[0], cd->nbas[1], cd->pr, ixsize);
    fflush(wr);
  }

#define dm_cand_vec_format "2006-02-19"

void dm_cand_vec_write(FILE *wr, dm_cand_vec_t *cdv)
  { 
    /* Find maximum abs of {id,name,index}: */
    int idmax = 999999; 
    int namemax = 10;
    int ixmax = 999999;
    /* FIX !!! should get max from {cdv} */
    int idsize = digits(idmax);
    int namesize = namemax;
    int ixsize = digits(ixmax);
    filefmt_write_header(wr, "dm_cand_vec", dm_cand_vec_format);
    fprintf(wr, "ncands = %d\n", cdv->nel);
    int i;
    for (i = 0; i < cdv->nel; i++)
      { dm_cand_write(wr, &(cdv->el[i]), idsize, namesize, ixsize);
        fputc('\n', wr);
      }
    filefmt_write_footer(wr, "dm_cand_vec");
    fflush(wr);
  }
  
void dm_cand_vec_write_named(dm_cand_vec_t *cdv, char *name, char *tag)
  { FILE *wr = dm_open_write(name, tag, ".cdv", TRUE);
    dm_cand_vec_write(wr, cdv);
    fclose(wr);
  }
 
void dm_cand_paint(dm_cand_t *cd, float_image_t *img, int scale)
  { /* Get sequence sizes {nx,ny}: */
    int nx = cd->nbas[0];
    int ny = cd->nbas[1];
    int ng = dm_pairing_num_rungs(cd->pr);
    /* Compute average score per rung: */
    double avsc = cd->score / ((double)ng);
    float v = (float)avsc;
    /* Paint rungs: */
    dm_pairing_paint(cd->pr, nx, ny, v, img, scale);
 }

void dm_cand_image_write_named(dm_cand_t *cd, char *name, char *tag)
  { /* Get sequence sizes {nx,ny}: */
    int nx = cd->nbas[0];
    int ny = cd->nbas[1];
    /* Choose image reduction factor {scale} and allocate the image {img}: */
    dm_image_t *img;
    int scale;
    dm_alloc_image(nx, ny, &img, &scale);
    /* Fill the image with "no value": */
    float_image_fill(img, (float)(-INF));
    /* Paint candidate on image: */
    dm_cand_paint(cd, img, scale);
    /* Normalize image and write it out: */
    dm_cand_write_painting_as_pgm(img, name, tag);
    /* Reclaim image space: */
    float_image_free(img);
  }
  
void dm_cand_vec_image_write_named
  ( dm_cand_vec_t *cdv,
    int nx, 
    bool_t circx,
    int ny,
    bool_t circy,
    char *name, 
    char *tag
  )
  {
    /* Choose image reduction factor {scale} and allocate the image {img}: */
    dm_image_t *img;
    int scale;
    dm_alloc_image(nx, ny, &img, &scale);
    /* Fill the image with "no value": */
    float_image_fill(img, (float)(-INF));
    /* Enumerate candidates for {xp,yp} and accumulate the pairs: */
    int ic; 
    for (ic = 0; ic < cdv->nel; ic++)
      { /* Get candidate with index {ic}: */
        dm_cand_t *cd = &(cdv->el[ic]);
        /* Check whether seqences have same size: */
        demand(cd->nbas[0] == nx, "inconsistent nx");
        demand(cd->nbas[1] == ny, "inconsistent ny");
        /* Paint it: */
        dm_cand_paint(cd, img, scale);
      }
    
    /* Normalize image and write it out: */
    dm_cand_write_painting_as_pgm(img, name, tag);
    /* Reclaim image space: */
    float_image_free(img);
  }
  
void dm_cand_write_painting_as_pgm(float_image_t *img, char *name, char *tag)
  { /* Find actual range of image (excluding infinite pixels): */
    double lov, hiv;
    dm_compute_min_max_image_range(img, &lov, &hiv);
    /* Adjust {lov} so that lowest score maps to gray: */
    lov = lov - (hiv - lov);
    /* Write out as PGM: background white, candidates gray to black: */
    dm_write_float_image_as_pnm(img, hiv, lov, name, tag);
  }
    
