#ifndef dm_cand_H
#define dm_cand_H

/* A candidate is a pairing with a score. */
/* Last edited on 2007-01-04 01:57:27 by stolfi */

#define dm_cand_H_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)"

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_rung.h>
#include <dm_pairing.h>

#include <vec.h>
#include <float_image.h>

#include <stdint.h>

typedef struct dm_cand_t
  { dm_seq_id_t id[2];   /* Internal identifiers of the two sequences. */
    char *name[2];       /* External identifiers of the two sequences. */
    int nbas[2];         /* Number of bases in each sequence. */
    bool_t circ[2];      /* Tells whether each sequence is circular. */
    dm_pairing_t *pr;    /* A partial pairing between the two sequences. */
    double score;        /* Value (goodness) of cand, for sorting. */
  } dm_cand_t;
  /* A candidate pairing between two sequences {seq[0],seq[1]},
    each described by the parameters {id[j],name[j],nbas[j],circ[j]}.
    
    Each rung of the pairing {pr} must consist of a valid index {c[0]}
    into {seq[0]}, and a valid index {c[1]} into the {seq[1]}. The
    first rung, in particular, must have {c[j]} in the range
    {0..nbas[j]-1} for both {j}. Moreover, if {circ[j]} is FALSE, this
    constraint must be satisfied for every rung.
    
    In the current implementation, the pairing must be unitary and
    monotonic. Note that if {circ[j]} is TRUE the pairing may have
    indices {c[j]} that exceed {nbas[j]}. Those indices should be
    taken modulo {nbas[j]} when indexing into {seq[j]}.
    
    The pairing {pr} may be circular. In that case one must have
    {circ[0]=circ[1]=TRUE}. Moreover, the coordinate period vector
    {per = dm_pairing_period(pr)} must have {per.c[j] % nbas[j] == 0}
    and {per.c[j] > 0} (i.e., it must make a positive number of full
    turns around {seq[j]}).
    
    A circular pairing may cover the same part of {seq[j]} more than
    once. This situation may be useful during multiscale search (to
    represent multiple pre-candidates between two circular sequences),
    and may even make sense for some applications. */

bool_t dm_cand_is_valid(dm_cand_t *cd);
  /* Runs consistency tests on the candidate {cd}. */

dm_cand_t dm_cand_from_pairing
  ( dm_seq_t *xp, 
    dm_seq_t *yp, 
    dm_pairing_t *pr,
    double score
  );
  /* Creates a candidate between the two sequences {xp,yp}
    with pairing {pr} and the given {score}.
    
    If the pairing {pr} is circular, each coordinate period must be
    positive and a multiple of the corresponding sequence length. The
    two sequences must be circular.
    
    If the pairing {pr} is open, each sequence must be either circular
    or long enough to contain the corresponding endpoint of the last
    rung of {pr}. */

vec_typedef(dm_cand_vec_t, dm_cand_vec, dm_cand_t);
  /* A {dm_cand_vec_t} is a vector of {dm_cand_t}. */

dm_cand_vec_t dm_cand_get_best_perfect
  ( dm_seq_t *xp,
    dm_seq_t *yp,
    dm_step_score_proc_t *step_score,
    int minBases, 
    double minScore,
    int maxCands,
    double frac
  );
  /* Build a set of initial cands between {*xp} and {*yp}. Considers
    only maximal perfect cands with at least {minBases} bases and
    score at least {minScore}. Returns only the best {maxCands} cands in
    order of increasing score, excluding candidates that are contained
    in better ones at the {frac} level of containment.
    
    The score of a pairing is taken to be be the sum of the
    {step_score} for all its steps, including loopback step
    if the pairing is circular, and none-to-first-rung and
    last-rung-to-none if it is open. */

int dm_cand_count_common_rungs(dm_cand_t *ca, dm_cand_t *cb);
  /* Number of rungs that are shared between {ca} and {cb}.
    The two candidates must refer to the same two sequences,
    otherwise the result is 0. Takes circularity of the 
    sequences into account. */

/* MAPPING CANDIDATES BETWEEN SCALES */

dm_cand_t dm_cand_map_to_finer(dm_cand_t *cd, int nx, int ny, int nwtb);
  /* Given a candidate {cdold} at some positive scale {r}, returns the
    corresponding candidate {cdnew} at scale {r-1}. Assumes that the
    two sequences being paired have {nx} and {ny} datums at
    the finer scale {r-1}, and that the filtering used a weight table
    of {nwtb} entries. The candidate's score is preserved.
    
    Let {ox,oy} be the lengths of the two sequences paired by {cd}.
    Every X coordinate {x} of a rung is mapped to {floor(x*(nx/ox) +
    dxy)}, and every Y coordinate {y} to {floor(y*(ny/oy) + dxy)}.
    Then these rungs are interpolated so as to produce a unitary
    monotonic mapping. */
 
dm_cand_vec_t dm_cand_vec_map_to_finer(dm_cand_vec_t *cdv, int nx, int ny, int nwtb);
  /* Applies {dm_cand_map_to_finer} to each element of the candidate vector 
    {cdv}, and returns a vector with the results. */

/* RANDOM CANDIDATES */

dm_cand_t dm_cand_throw
  ( dm_seq_id_t ida, 
    char *namea, 
    int na, 
    bool_t circa, 
    dm_seq_id_t idb, 
    char *nameb, 
    int nb, 
    bool_t circb,
    int len,
    bool_t circp,
    bool_t diagonal,
    double skipProb
  );
  /* Generates a random candidate between two hypothetical datum
    sequences. The first sequence is assumed to have numeric
    identifier {ida}, name {namea}, datum count {na}, and circularity
    {circa}. Ditto for {idb,nameb,nb,circb} and the other sequence.
    
    The pairing will cover approximately {len} bases on each sequence
    (it may vary somewhat, depending on other constraints).
    If either sequence is open, {len} must not exceed the
    corresponding sample count ({na} or {nb}).
    
    The pairing will be circular iff {circp} is TRUE. In that case,
    the {len} parameter is ignored; both sequences must be circular,
    and each may be covered multiple times.
    
    If {diagonal} is TRUE, the endpoints are chosen near
    the diagonal of the rectangle {{0..nx-1}{0..ny-1}}. 
    The parameter {skipProb} specifies the approximate ratio
    of axial steps (`skips') to diagonal steps (`hops'). */

dm_cand_vec_t dm_cand_vec_throw
  ( int ncd,
    dm_seq_id_t ida, 
    char *namea, 
    int na, 
    bool_t circa, 
    dm_seq_id_t idb, 
    char *nameb, 
    int nb, 
    bool_t circb,
    int minlen,
    int maxlen,
    double circProb,
    double diagProb,
    double skipProb
  );
  /* Generates a vector of {ncd} random candidates between two
    sequences. 
    
    The candidate lengths are chosen in the range {minlen..maxlen}
    with uniform probability.
    
    If both sequences are circular, the candidates will have circular
    pairings with probebility {circProb}; in that case {maxlen} is
    ignored. Otherwise all candidates will have open pairings.
    
    With probability {diagProb}, the pairings will start and end near
    the diagonal of the rectangle {{0..nx-1}{0..ny-1}} 
    
    The other parameters are as explained under {dm_cand_throw}. */


/* CANDIDATE I/O */

void dm_cand_read(FILE *rd, dm_cand_t *cd);
  /* Reads a candidate from {rd} and stores it in {*cd}. The data should start
    with the id, name, and length of the first sequence, followed by the id, name,
    and length of the second sequence, followed by the candidate's score, followed by a 
    a description of the pairing (see {dm_pairing_read}). These fields must be 
    separated by whitespace. The names should not contain embedded blanks. */

void dm_cand_debug(char *tag, dm_cand_t *cd);
  /* Prints the {tag} and the candidate {cd} to {stderr}. */
    
void dm_cand_write(FILE *wr, dm_cand_t *cd, int idsize, int namesize, int ixsize);
  /* Writes the candidate {cd} to the file {wr}.
    The numeric IDs of sequences are written with {idsize} digits, 
    their names are written with {namesize} chars, and 
    the datum indices are written with {ixsize} digits. */

void dm_cand_vec_write(FILE *wr, dm_cand_vec_t *cdv);
  /* Writes the list of candidates {cdv} to the file {wr}. */

void dm_cand_vec_write_named(dm_cand_vec_t *cdv, char *name, char *tag);
  /* Writes the list of pairings {cdv} to the file named "{name}{tag}.cdv". */

/* PLOTTING CANDIDATES */

void dm_cand_paint(dm_cand_t *cd, float_image_t *img, int scale);
  /* Paints the candidate {cd}, reduced by {1/scale}, on
    image {img}.  The color {v} is the candidate's score.
    
    More precisely, each rung {ix,iy} of {cd->p} is reduced to the
    rectangle {{0..nx-1}{0..ny-1}}, where {nx} and {ny} are the
    sequence lengths; and the value {v} is added to pixel in column
    {ix/scale} and row {iy/scale} of {img}. */

void dm_cand_image_write_named
  ( dm_cand_t *cd,
    char *name, 
    char *tag
  );
  /* Writes to disk a PGM image called "{name}{tag}.pgm" showing the
    candidate {cd}. Assumes that the candidate refers to
    two sequences with datum counts {nx,ny} and circularities
    {circx,circy}. 
    
    A pixel {(ix,iy)} is marked if the candidate pairs sample {ix} of
    sequence 0 with sample sample {iy} of sequence 1. The darkness of
    the pixel is proportional to the score of the candidate divided by
    its rung count, times how many rungs of the pairing got mapped to
    that pixel. The image is scaled down as needed to keep its size
    below some reasonable maximum size. */

void dm_cand_vec_image_write_named
  ( dm_cand_vec_t *cdv,
    int nx, 
    bool_t circx,
    int ny,
    bool_t circy,
    char *name, 
    char *tag
  );
  /* Writes to disk a PGM image called "{name}{tag}.pgm" showing the
    candidates of {cdv}.   All candidates must refer to the same
    two sequences, which should have {nx,ny} samples and circularities
    {circx,circy}, respectively.  The image is similar to that produced by 
    {dm_cand_paint}, summed over all candidates {cd} in {cdv}. */

#endif
