/* See {dm_image.h} */
/* Last edited on 2007-01-04 00:11:38 by stolfi */

#define dm_image_C_COPYRIGHT \
  "Copyright  2005 by the State University of Campinas (UNICAMP)"

#include <dm_basic.h>

#include <dm_image.h>

#include <affirm.h>
#include <float_image.h>
#include <float_pnm_image.h>
#include <jspnm_image.h>

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

void dm_alloc_image(int nx, int ny, dm_image_t **imgp, int *scalep)
  {
    /* Decide image downscaling factor {scale}: */
    int nbig = (nx > ny ? nx : ny); /* Length of largest sequence. */
    int nmax = 1024; /* Maximum alowed image size. */
    int scale = (nbig + nmax - 1)/nmax;
    if (scale == 0) { scale = 1; }
    /* Compute actual image size {npx,npy}: */
    int npx = (nx - 1)/scale + 1;
    int npy = (ny - 1)/scale + 1;
    
    /* Allocate image: */
    dm_image_t *img = float_image_new(1, npx, npy);
    
    /* Return results: */
    (*scalep) = scale;
    (*imgp) = img;
  }

void dm_compute_avg_dev_image_range(dm_image_t *img, double *lo, double *hi, double ns)
  { demand(img->sz[0] == 1, "image is not grayscale");
    int nx = img->sz[1];
    int ny = img->sz[2];

    /* Compute the mean image value {avg}: */
    double sum = 0.0;
    int tot = 0;
    int ix, iy;
    for (iy = 0; iy < ny; iy++)
      { for (ix = 0; ix < nx; ix++)
          { float fv = float_image_get_sample(img, 0, ix, iy);
            if (fabs(fv) != (float)INF) { sum += fv; tot++; }
          }
      }
    double avg = (tot == 0 ? 0.0 : sum/tot);

    /* Compute the standard deviation {dev}, make sure it is positive: */
    sum = 0;
    for (iy = 0; iy < ny; iy++)
      { for (ix = 0; ix < nx; ix++)
          { float fv = float_image_get_sample(img, 0, ix, iy);
            if (fabs(fv) != (float)INF)
              { double dv = fv - avg; 
                sum += dv*dv;
              }
          }
      }
    double dev = (tot <= 1 ? 0.0 : sqrt(sum/(tot-1)));

    /* Choose black and white values: */
    double vlo = avg - ns*dev;
    double vhi = avg + ns*dev;
    assert(vlo <= vhi);
    if (vlo == vhi)
      { vlo = nextafter(vlo, -INF);
        vhi = nextafter(vhi, +INF);
      }

    /* Return them: */
    (*lo) = vlo; (*hi) = vhi;
  }

void dm_compute_min_max_image_range(dm_image_t *img, double *lo, double *hi)
  { demand(img->sz[0] == 1, "image is not grayscale");
    int nx = img->sz[1];
    int ny = img->sz[2];

    /* Compute the extremal values {vmin,vmax} of finite pixels: */
    double vmin = +INF, vmax = -INF;
    int ix, iy;
    for (iy = 0; iy < ny; iy++)
      { for (ix = 0; ix < nx; ix++)
          { float fv = float_image_get_sample(img, 0, ix, iy);
            if (fabs(fv) != (float)INF) 
              { if (fv < vmin) { vmin = fv; }
                if (fv > vmax) { vmax = fv; }
              }
          }
      }

    /* Provide default range {[0_0]} if there are no finite pixels: */
    if (vmin > vmax) { vmin = vmax = 0; }
    
    /* Make sure that {vmin < vmax}: */
    if (vmin == vmax) 
      { vmin = nextafter(vmin, -INF);
        vmax = nextafter(vmax, +INF);
      }

    /* Return them: */
    (*lo) = vmin; (*hi) = vmax;
  }

float_image_t *dm_image_colorize_float_image(dm_image_t *gim, double maxf)
  { /* Check the channel count and get image dimensions: */
    demand(gim->sz[0] == 1, "the input image must be single-channel");
    int NX = gim->sz[1]; /* Number of columns. */
    int NY = gim->sz[2]; /* Number of rows. */
    int NC = 3; /* Number of channels in output image. */
    
    /* Limiting colors: */
    float zer[3] = { 1.000, 1.000, 1.000 }; /* Lum = 1.000; for {v = 0}. */
    float pos[3] = { 1.000, 0.333, 0.000 }; /* Lum = 0.500; for {v=+maxf}. */
    float neg[3] = { 0.000, 0.667, 1.000 }; /* Lum = 0.500; for {v=-maxf}. */

    /* Allocate PGM/PPM image: */
    float_image_t *cim = float_image_new(NC, NX, NY);
    
    /* Convert monochrome pixels from {gim} to color pixels of {cim}: */
    int x, y, c;
    for(y = 0; y < NY; y++)
      { for(x = 0; x < NX; x++)
          { /* Get pixel {v} from monochrome image: */
            float v = float_image_get_sample(gim, 0, x, y);
            /* Compute the corresponding color {clr}: */
            float clr[NC];
            if (fabsf(v) == fabsf(INF))
              { /* Assume pixel as undefined, colorize it with black: */
                for (c = 0; c < NC; c++) { clr[c] = 0; }
              }
            else
              { /* Choose the limiting color {lim} appropriate to the sign of {v}: */
                float *lim = (v >= 0 ? pos : neg); /* Limiting color for sign of {v}. */
                /* Compute intensity {s} of {v} relative to {maxf}, and its complement {t}: */
                double s = fabs(v)/maxf;
                assert(s >= 0); 
                if (s > 1) { s = 1; }
                double t = 1 - s;
                /* Interpolate with {s} between {zer} and {lim} to get the color {clr}: */
                for (c = 0; c < NC; c++) { clr[c] = t*zer[c] + s*lim[c]; }
              }
            /* Store {clr} in color image: */
            for (c = 0; c < NC; c++) 
              { float_image_set_sample(cim, c, x, y, clr[c]); }
          }
      }
    return cim;
  }

void dm_write_float_image_as_pnm
  ( dm_image_t *img, 
    double minv, 
    double maxv, 
    char *name,
    char *tag
  )
  { int NC = img->sz[0];
    demand((NC == 1) || (NC == 3), "bad number of channels"); 
    /* Prepare the per-channel {lo,hi} parameters from single {minv,maxv}: */
    double lov[NC];
    double hiv[NC];
    int c; 
    for (c = 0; c < NC; c++) { lov[c] = minv; hiv[c] = maxv; }
    /* Convert the float image {img} to an integer image {pnm}: */
    pnm_image_t *pnm = float_image_to_pnm_image(img, NC, NULL, lov, hiv, 255, TRUE);
    /* Construct the file name {fileName}: */
    char *ext = (NC == 1 ? ".pgm" : ".ppm");
    /* Open the output file {wr} and write {pnm} there: */
    FILE *wr = dm_open_write(name, tag, ext, TRUE);
    pnm_image_fwrite(wr, pnm, FALSE);
    /* Cleanup: */
    fclose(wr);
    pnm_image_free(pnm);
  }

void dm_bump_image(dm_image_t *img, int ix, int iy, int scale, float v)
  { int ipx = ix/scale;
    int ipy = iy/scale;
    float *p = float_image_get_sample_address(img, 0, ipx, ipy);
    if (fabs(*p) == (float)INF)
      { (*p) = v; }
    else
      { (*p) += v; }
  }

