#define PROG_NAME "dm_match"
#define PROG_DESC "multiscale matching of two DNA/RNA strings"
#define PROG_VERS "1.0"

/* Last edited on 2007-04-06 20:49:42 by stolfi */

#define dm_match_C_COPYRIGHT \
  "Copyright  2005 by the State University of Campinas (UNICAMP) " \
  " and the Federal Fluminense University (UFF)"

#define PROG_HELP \
  PROG_NAME " \\\n" \
  "  [ -maxLevel {MAX_LEVEL} ] \\\n" \
  "  -initFilter " dm_weight_args_HELP " \\\n" \
  "  -incrFilter " dm_weight_args_HELP " \\\n" \
  "  -scores " dm_score_args_HELP " \\\n" \
  "  [ -delta {DELTA} ] [ -kappa {KAPPA} ] \\\n" \
  "  [ -contFrac {FRAC} ] \\\n" \
  "  -maxMatches {MAX_MATCHES} \\\n" \
  "  -minBases {MIN_BASES} \\\n" \
  "  {SEQ_FILE_A} {CIRC_A} \\\n" \
  "  {SEQ_FILE_B} {CIRC_B} \\\n" \
  "  {OUT_NAME}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  "  " PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  This program finds matching subsegments of two given" \
  " DNA/RNA sequence files, {SEQ_FILE_A} and {SEQ_FILE_B}, by" \
  " multiscale dynamic programming with smooth filtering.\n" \
  "\n" \
  "  The two input files should be in \".seq\" format" \
  " (see the INPUT FILE FORMAT section below).  The" \
  " arguments {CIRC_A} and {CIRC_B} are either \"T\" (true)" \
  " or \"F\" (false).  If {CIRC_A} is true, the" \
  " sequence in {SEQ_FILE_A} is assuemd to be circular (like a bacterial" \
  " chromosome or plasmid).  Ditto for {CIRC_B} and {SEQ_FILE_B}.\n" \
  "\n" \
  "INPUT FILE FORMAT\n" \
  dm_nucleic_file_format_INFO "\n" \
  "\n" \
  "OUTPUT FILES\n" \
  "  All output files will have names starting with {OUT_NAME}.\n" \
  "\n" \
  "OPTIONS\n" \
  "  -maxLevel {MAX_LEVEL} \n" \
  "    This optional argument specifies the maximum number of filtering steps" \
  " to use.  The actual maximum level may be less, depending on" \
  " the {MIN_BASES} parameter.  The default is the maximum level" \
  " implied by {MIN_BASES}.\n" \
  "\n" \
  "  -maxMatches {MAX_MATCHES} \n" \
  "    This mandatory argument specifies the maximum number of matched" \
  " subsequence pairs to produce at the original scale.  (The" \
  " limit is multiplied by {2^r} at scale {r}.)\n" \
  "\n" \
  "  -minBases {MIN_BASES} \n" \
  "    This mandatory argument specifies the minimum number of bases" \
  " in the output matches.\n" \
  "\n" \
  "  -initFilter " dm_weight_args_HELP " \n" \
  "  -incrFilter " dm_weight_args_HELP " \n" \
  "    These mandatory arguments specify the" \
  " weights of the filter to be used at the" \
  " first filtering step and at subsequent" \
  " filtering steps.  " dm_weight_args_norm_sum_INFO "\n" \
  "\n" \
  "  -delta {DELTA}\n" \
  "    This optional argument specifies the" \
  " amount of adjustment allowed for the X and Y coordinates of" \
  " internal rungs of each pairing.  The default is 3.\n" \
  "\n" \
  "  -kappa {DELTA}\n" \
  "    This optional argument specifies the" \
  " amount of X and Y extension allowed" \
  " at either end of each pairing.  The default is 6.\n" \
  "\n" \
  "  -contFrac {FRAC} \n" \
  "    This optional argument specifies a fractional" \
  " similarity level for discarding similar" \
  " candidates.  If specified and" \
  " positive, the program discards (at each scale) any" \
  " candidate {A} if there is a candidate {B} that" \
  " \"mostly contains\" {A} and has better score" \
  " than {A}.  By \"mostly contains\" we mean that the" \
  " fraction of {A} pairs contained in {B} is {FRAC}" \
  " or more.  For example, if {FRAC} is 0.95, then" \
  " {B} \"mostly contains\" {A} if 95% or more of" \
  " the pairs in {A} are also in {B}.  The default" \
  " is \"-contFrac 0.99\".\n" \
  "\n" \
  "  -scores " dm_score_args_HELP "\n" \
  "    This mandatory argument specifies the" \
  " scoring criterion to be used when evaluating" \
  " and comparing candidates.  " dm_score_args_INFO "\n" \
  "\n" \
  argparser_help_info_HELP_INFO "\n" \
  "SEE ALSO\n" \
  "  dm_test_filter(1)\n" \
  "\n" \
  "AUTHOR\n" \
  "  This program was created in jun/2005 by J. Stolfi, based" \
  " on the fragment reassembly programs of Helena C. G. Leito's" \
  " Ph.D. thesis.\n" \
  "\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " dm_match_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS  

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_weight.h>
#include <dm_nucleic.h>
#include <dm_rung.h>
#include <dm_pairing.h>
#include <dm_score.h>
#include <dm_cand.h>
#include <dm_multi.h>
#include <dm_image.h>

#include <affirm.h>
#include <float_image.h>
#include <float_pnm_image.h>
#include <jspnm_image.h>
#include <argparser.h>

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

typedef struct dm_options_t 
  { char *seqNameA;      /* Name of file with sequence {A} (minus ".bas" ext). */
    bool_t circA;        /* TRUE iff sequence {A} is circular. */
    char *seqNameB;      /* Name of file with sequence {B} (minus ".gdb" ext). */
    bool_t circB;        /* TRUE iff sequence {B} is circular. */
    /* Filtering parameters: */
    int maxLevel;        /* Maximum filtering level. */
    int minBases;        /* Min number of bases in match. */
    double_vec_t w0;     /* Filter weights for first stage (level 0 to level 1). */
    double_vec_t w1;     /* Filter weights for subsequent stages. */
    /* Multiscale matching parameters: */
    int maxMatches;      /* Max number o matches to report. */
    bool_t dontRefine;   /* Do not refine the candidates after mapping (debug). */
    double contFrac;     /* Containment fraction for discarding candidates. */
    /* Scoring parameters: */
    int delta;           /* Half-width of tableau around original pairing. */
    int kappa;           /* Extension of tableau beyond ends of pairing. */
    dm_score_args_t sc;  /* Candidate scoring parameters. */
    /* Output parameters: */
    char *outName;       /* Output file name prefix (minus extensions). */
  } dm_options_t;
  
int main(int argc, char**argv);

dm_options_t *dm_get_options(int argc, char**argv);
  /* Parses the command line options, packs 
    them into a {dm_options_t} record. */
 
dm_cand_vec_t dm_match_get_initial_cands
  ( dm_seq_t *a, 
    dm_seq_t *b,
    int maxMatches,
    double contFrac,
    int minBases,
    dm_step_score_proc_t *step_score
  );
  /* Given two sequences {a,b}, finds an initial set of up to
    {maxMatches} candidates, each covering at least {minBases}
    elements of both sequences. */

int main(int argc, char**argv)
  { 
    dm_options_t *o = dm_get_options(argc, argv);

    fprintf(stderr, "reading input sequences ...\n");
    dm_seq_t a = dm_seq_read_from_nucleic_file(0, "a", o->seqNameA, o->circA);
    dm_seq_t b = dm_seq_read_from_nucleic_file(1, "b", o->seqNameB, o->circB);

    fprintf(stderr, "displaying weight tables ...\n");
    char *wname0 = dm_weight_make_descr(&(o->w0), "%7.5f");
    dm_weight_table_print(stderr, "initial", &(o->w0));
    char *wname1 = dm_weight_make_descr(&(o->w1), "%7.5f");
    dm_weight_table_print(stderr, "incremental", &(o->w1));
  
    fprintf(stderr, "filtering to scales %d .. %d ...\n", 0, o->maxLevel);
    dm_seq_t ar[o->maxLevel + 1];
    dm_seq_t br[o->maxLevel + 1];
    dm_multi_seq_filter(&a, o->maxLevel, &(o->w0), wname0, &(o->w1), wname1, ar);
    dm_multi_seq_filter(&b, o->maxLevel, &(o->w0), wname0, &(o->w1), wname1, br);
    
    /* Find the maximum effective level {M}: */
    int M = o->maxLevel;
    while ((M > 0) && ((dm_seq_length(&(ar[M])) == 0) || (dm_seq_length(&(br[M])) == 0)))
      { M--; }
    fprintf(stderr, "useful scales %d .. %d ...\n", 0, M);
    
    /* Grab coarsest seqs {aini,bini}, lengths {naini,nbini}, flags {acirc,bcirc}: */
    dm_seq_t *aini = &(ar[M]);
    dm_seq_t *bini = &(br[M]);
    int naini = dm_seq_length(aini); bool_t acirc = dm_seq_is_circular(aini);
    int nbini = dm_seq_length(bini); bool_t bcirc = dm_seq_is_circular(bini);

    auto double step_score(dm_rung_t g, dm_rung_t h, dm_seq_t *xp, dm_seq_t *yp);

    double step_score(dm_rung_t g, dm_rung_t h, dm_seq_t *xp, dm_seq_t *yp)
      { return dm_score_step(&(o->sc), g, h, xp, yp); }

    fprintf(stderr, "plotting distance matrix for initial seq pair ...\n");
    dm_pairing_seq_seq_image_write_named(aini, bini, &step_score, o->outName, "-ini-eq");
    
    fprintf(stderr, "finding initial candidates ...\n");
    int S = (1 << M);
    int maxCands = o->maxMatches*S;
    int minLength = (o->minBases + S - 1)/S;
    dm_cand_vec_t cdvini = dm_match_get_initial_cands
      (aini, bini, maxCands, o->contFrac, minLength, &step_score);

    fprintf(stderr, "writing initial candidates ...\n");
    dm_cand_vec_write_named(&cdvini, o->outName, "-ini-cd");
    dm_cand_vec_image_write_named(&cdvini, naini, acirc, nbini, bcirc, o->outName, "-ini-cd");
      
    fprintf(stderr, "multiscale mapping and refining ...\n");
    dm_cand_vec_t cdvfin = dm_multi_find_matches
      ( &cdvini,
        ar, br, 
        o->maxMatches, M, o->minBases, 
        o->w0.nel, o->w1.nel,
        o->contFrac, (! o->dontRefine),
        o->delta, o->kappa, TRUE,
        &step_score,
        o->outName
      );

    dm_multi_seq_free(ar, o->maxLevel);
    dm_multi_seq_free(br, o->maxLevel);
    
    fprintf(stderr, "plotting pair-to-pair distances ...\n");
    dm_seq_t *afin = &(ar[0]); assert(acirc == dm_seq_is_circular(afin));
    dm_seq_t *bfin = &(br[0]); assert(bcirc == dm_seq_is_circular(bfin));
    int nafin = dm_seq_length(afin);
    int nbfin = dm_seq_length(bfin);
    dm_pairing_seq_seq_image_write_named(afin, bfin, &step_score, o->outName, "-fin-eq");
    
    fprintf(stderr, "writing final candidates ...\n");
    dm_cand_vec_write_named(&cdvfin, o->outName, "-fin-cd");
    dm_cand_vec_image_write_named(&cdvfin, nafin, acirc, nbfin, bcirc, o->outName, "-fin-cd");

    return 0;
  }
    
dm_cand_vec_t dm_match_get_initial_cands
  ( dm_seq_t *a, 
    dm_seq_t *b,
    int maxCands,
    double contFrac,
    int minLength,
    dm_step_score_proc_t *step_score
  )
  { double minScore = 0.0;
    dm_cand_vec_t cdv = dm_cand_get_best_perfect
      ( a, b, step_score,
        minLength, minScore, maxCands, contFrac
      );
    fprintf(stderr, "found %d candidates at coarsest level\n", cdv.nel);
    return cdv;
  }

dm_options_t *dm_get_options(int argc, char**argv)
  { 
    dm_options_t *o = (dm_options_t *)notnull(malloc(sizeof(dm_options_t)), "no mem");
    
    argparser_t *pp = argparser_new(stderr, argc, argv);
    argparser_set_help(pp, PROG_HELP);
    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);
    
    /* Note that level 40 requires seqs with at least {2^40} bases: */ 
    if (argparser_keyword_present(pp, "-maxLevel"))
      { o->maxLevel = argparser_get_next_int(pp, 0, 40); }
    else
      { o->maxLevel = 40; } 
    
    argparser_get_keyword(pp, "-minBases");
    o->minBases = argparser_get_next_int(pp, 0, INT_MAX);
    
    argparser_get_keyword(pp, "-scores");
    o->sc = dm_score_args_parse(pp);
    
    if (argparser_keyword_present(pp, "-delta"))
      { o->delta = argparser_get_next_int(pp, 0, INT_MAX); }
    else
      { o->delta = 3; }
    
    if (argparser_keyword_present(pp, "-kappa"))
      { o->kappa = argparser_get_next_int(pp, 0, INT_MAX); }
    else
      { o->kappa = 6; }
    
    argparser_get_keyword(pp, "-maxMatches");
    o->maxMatches = argparser_get_next_int(pp, 0, INT_MAX);
    
    argparser_get_keyword(pp, "-initFilter");
    o->w0 = dm_weight_args_parse(pp, TRUE);

    argparser_get_keyword(pp, "-incrFilter");
    o->w1 = dm_weight_args_parse(pp, TRUE);

    if (argparser_keyword_present(pp, "-contFrac"))
      { o->contFrac = argparser_get_next_double(pp, -1.0, +1.0); }
    else
      { o->contFrac = 0.99; }
    
    argparser_skip_parsed(pp);
    
    o->seqNameA = argparser_get_next(pp);
    o->circA = argparser_get_next_bool(pp);
    
    o->seqNameB = argparser_get_next(pp);
    o->circB = argparser_get_next_bool(pp);
    
    o->outName = argparser_get_next(pp);
    
    o->dontRefine = FALSE; /* DEBUG */
    
    argparser_finish(pp);
    
    return o;
  }
