#ifndef dm_pairing_H
#define dm_pairing_H

/* Monotonic pairings between elements of a sequences */
/* Last edited on 2007-01-04 01:54:48 by stolfi */

#define dm_pairing_H_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)" \
  " and the Federal Fluminense University (UFF)"

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_rung.h>

#include <sign.h>
#include <vec.h>
#include <float_image.h>

#include <stdint.h>

/* PAIRINGS

  A `pairing' is a sequence of one or more rungs {p[i]}, for some set
  of integer `rung indices' {i}.
  
  There are two basic kinds of pairings: `open' and `circular'. 
  
  If {p} is an open pairing, then {p[i]} is defined only for {i} in
  {0..ng-1}, for some integer {ng >= 0}.
  
  If {p} is a circular pairing, then {p[i]} is defined for any integer
  {i}. However, the rungs are `integral-periodic', meaning that there
  is a positive integer {ng} (the `rung index period') and integers
  {DX,DY} (the `coordinate periods') such that {p[i+ng] =
  p[i]+(DX,DY)} for every {i}. */
  
typedef struct dm_pairing_t dm_pairing_t;

int dm_pairing_num_rungs(dm_pairing_t *p);
  /* Returns the number {ng} of rungs in the pairing {p}. 
    If {p} is circular, the result is the rung index period. */

bool_t dm_pairing_is_circular(dm_pairing_t *p);
  /* Returns TRUE iff {p} is a circular pairing. */

dm_rung_t dm_pairing_period(dm_pairing_t *p);
  /* Returns the coordinate periods {PX,PY} of {p}, packed as a rung.
    If {p} is open, the result is {(0,0)}. */

bool_t dm_pairing_is_valid (dm_pairing_t *p, bool_t die);
  /* Runs consistency checks on the pairing {p}.
    If {die} is true, and the pairing is not valid,
    fails with an error message.  If {die} is FALSE,
    returns TRUE or FALSE normally without any messages. */

dm_rung_t dm_pairing_get_rung(dm_pairing_t *p, int i);
  /* Obtains the rung with index {i} in the pairing {*p}. If {p} is
    open, the index {i} must be in the range {0..ng-1}, where {ng} is
    the number of rungs. If {p} is circular, the index {i} can be any
    integer. */

dm_rung_vec_t dm_pairing_collect_rungs(dm_pairing_t *p, int nx, int ny);
  /* Returns the set of all distinct fundamental rungs of {p},
    namely rungs {p[0..ng-1]} where {ng = dm_pairing_num_rungs(p)}.
    
    The parameters {nx} and {ny} must be non-negative. If {nx} is
    positive, the X coordinates of every rung are reduced modulo {nx}.
    (If {p} is periodic, this reduction makes sense only if its X period
    is a multiple of {nx}.)  Ditto for {ny} and the Y coordinates.
    
    The procedure sorts the list of reduced rungs by {dm_rung_lex_compare},
    and discards duplicated rungs. */

int dm_pairing_count_common_rungs(dm_pairing_t *pa, dm_pairing_t *pb, int nx, int ny);
  /* Given two pairings {pa,pb}, returns the number of rungs that are
    shared between them. If {pa} is periodic, considers only the rungs
    in the fundamental period. Ditto for {pb}.
    
    The parameters {nx} and {ny} must be non-negative. If {nx} is
    non-zero, the X coordinates of every rung are reduced modulo {nx}.
    (If {p} is periodic, this reduction makes sense only if its X
    period is a multiple of {nx}.) Ditto for {ny} and the Y
    coordinates.
    
    In other words, this procedure computes the size of the
    intersection of the two sets {dm_pairing_collect_rungs(pa,nx,ny)}
    and {dm_pairing_collect_rungs(pb,nx,ny)}. */

dm_pairing_t *dm_pairing_from_rung_vec(dm_rung_vec_t *gv, bool_t circp);
  /* Creates a pairing from a rung vector {gv}.
    
    If {circp} is false, the pairing will be an open one, consisting of
    the rungs in {gv} only. If {circp} is true, the pairing will be
    circular, with rung index period {ng = gv.nel-1}, and the 
    period will be assumed to be {gv[ng] - gv[0]}. */

dm_pairing_t *dm_pairing_perfect(dm_rung_t *gini, int ng, bool_t circp);
  /* Creates a pairing with rungs {gini + (k,k)} for {k} in {0..ng-1}.
    
    If {circp} is false, the pairing will be an open one, with those
    {ng} rungs only. If {circp} is true, the pairing will be circular,
    with rung index period {ng} and coordinate periods {(ng,ng)}. */

int64_t dm_pairing_perfect_max_length(int nx, bool_t circx, int ny, bool_t circy);
  /* Maximum length of a perfect pairing between two sequences
    {x,y} with {nx} and {ny} bases, with circularities {circx} and
    {circy}, respectively. */

typedef void dm_pairing_use_proc_t(int ix, int iy, int64_t ng, bool_t circp);
  /* A client procedure that processes a perfect pairing between two sequences.
    The pairing consists of pairs {(ix+k,iy+k)} for {k} in {0..ng-1},
    where each index is understood to be taken modulo the length of the 
    respective sequence. If {circp} is TRUE, the pairing itself should
    be assumed to be circular. */

void dm_pairing_enum_alignments
  ( int nx,       /* Number of bases on first sequence. */
    bool_t circx, /* TRUE if first sequence is circular. */
    int ny,       /* Number of bases on second sequence. */ 
    bool_t circy, /* TRUE if second sequence is circular. */ 
    dm_pairing_use_proc_t *use
  );
  /* Enumerates all possible maximal perfect pairings of two sequences
    {x,y} with {nx} and {ny} bases, with circularities {circx} and
    {circy}, respectively. Calls {use(ix, nx, iy, ny, ng, circp)} on
    each alignment.
    
    If both sequences are circular, each pairing itself will be
    circular, and will run {nx/g} times around {x}, and {ny/g} times
    around {y}, for a total of {nx*ny/g} sample pairs, where
    {g=gcd(nx,ny)}. If only {x} is circular, each pairing will cover
    {y} completely, and run {nx/ny} times around {x}. The analogous
    thing holds if only {y} is circular. If neither sequence is
    circular, each pairing will match a prefix of {x} with a suffix of
    {y}, or vice-versa, or the whole of {x} with an internal segment
    of {y}, or vice-versa. */
    
/* STEPS
  
  A `step' in a pairing is a pair of consecutive rungs.
  
  An open pairing {p} with {ng} rungs has {ng-1} steps 
  {p[i] -> p[i+1]} for {i} in {0..ng-2}.
  
  A circular pairing {p} with rung index period {ng} 
  has infintely many steps, but {ng} `fundamental steps'
  {p[i] -> p[i+1]} for {i} in {0..ng-1}. */
  
int dm_pairing_num_steps(dm_pairing_t *p);
  /* Returns the number steps in the pairing {p}. If {p} is circular,
    the result is the number of fundamental steps. */

typedef double dm_step_score_proc_t
  ( dm_rung_t g0, 
    dm_rung_t g1,
    dm_seq_t *xp, 
    dm_seq_t *yp
  ); 
  /* A procedure that computes a numerical score for a single step of a
    pairing between sequences {xp,yp}, between rung {g0} and rung
    {g1}. */

double dm_pairing_perfect_score
  ( dm_seq_t *xp,  /* A sequence . */
    int ix,        /* Starting index in sequence {xp}. */
    dm_seq_t *yp,  /* Another sequence. */
    int iy,        /* Starting index in sequence {yp}. */
    int ng,        /* Number of pairs. */
    bool_t circ,   /* TRUE if pairing should be considered circular. */
    dm_step_score_proc_t *step_score   /* Step scoring function. */
  );
  /* Computes the score for a perfect pairing between sequences 
    {xp,yp}. The pairing consists of rungs {(ix+k,iy+k)} 
    for {k = 0..ng-1}, modulo the lengths of the respective sequences,
    and the intervening steps. If {circ}, the pairing is assumed to include
    an additional step from the last rung to the first one.
    
    The score is the sum of {step_score} over all steps, including
    loopback step if the pairing is circular, and none-to-first-rung
    and last-rung-to-none if it is open. */

/* COINCIDENCE-MAP IMAGE */

void dm_pairing_seq_seq_image_write_named
  ( dm_seq_t *xp, 
    dm_seq_t *yp, 
    dm_step_score_proc_t *step_score, 
    char *name, 
    char *tag
  );
  /* Writes to disk a PPM image called "{name}{tag}.ppm" where each
    pixel {(ix,iy)} indicates the incremental cost of pairing 
    datum {ix} of sequence {xp} and datum {iy} of sequence {yp}.
    
    More precisely, the pixel value is the maximum cost of any
    valid pairing step that ends at rung {ix,iy}, including the step
    from nowhere.  
    
    Positive scores are redish, negative scores are bluish, zero
    scores are white. The image is scaled down as needed to keep its
    size below some reasonable maximum size. */

/* MAPPING PAIRINGS BETWEEN SCALES */

dm_pairing_t *dm_pairing_map_coordinates(dm_pairing_t *p, dm_rung_map_proc_t *map);
  /* Given a pairing {p}, returns a new pairing {q}
    such that {q[i] = map(p[i])} for all valid {i}.
    
    If {p} is open, {q} is open and has the same rung count. If {p} is
    circular, {q} is circular and has the same rung index period. */
    
dm_pairing_t *dm_pairing_map_to_finer
  ( dm_pairing_t *p, 
    int nxold, 
    int nyold, 
    int nxnew, 
    int nynew, 
    int nwtb
  );
  /* Applies {dm_seq_index_map_to_finer(ix,nxold,nxnew,nwtb)} to every
    X-coordinate {ix} in {p}, and {dm-seq_index_map_to_finer(iy,nyold,nynew,nwtb)} to every
    Y-coordinate {iy}. If {p} is monotonic, the result is monotonic. */

dm_pairing_t *dm_pairing_map_to_coarser
  ( dm_pairing_t *p, 
    int nxold, 
    int nyold, 
    int nxnew, 
    int nynew, 
    int nwtb
  );
  /* Applies {dm_seq_index_map_to_coarser(ix,nxold,nxnew,nwtb)} to
    every X-coordinate {ix} in {p}, and
    {dm-seq_index_map_to_coarser(iy,nyold,nynew,nwtb)} to every
    Y-coordinate {iy}. If {p} is monotonic, the result should be
    monotonic, except possibly for null steps. */

/* MONOTONIC AND UNITARY PAIRINGS
  
  A pairing, open or circular, is /monotonic/ if all its steps are
  monotonic. Thus, a monotonic pairing can be visualized as a finite
  or infinite path in the integer plane, whose steps are nonzero and
  directed into the first quadrant (including its boundary). */

bool_t dm_pairing_is_unitary_monotonic(dm_pairing_t *p);
  /* TRUE iff all steps in pairing {p} are unitary and monotonic. */

dm_pairing_t *dm_pairing_interpolate(dm_pairing_t *p);
  /* Given a monotonic pairing {p}, returns a minimal *unitary*
    monotonic pairing {q} that contains the rungs of {p} as a subset.
    This is achieved by connecting successive rungs by a 
    Bresenham-like path of diagonal and axial unitary steps.  */

/* RANDOM PAIRINGS */

dm_pairing_t *dm_pairing_throw
  ( int nx,
    bool_t circx, 
    int ny, 
    bool_t circy, 
    int len,
    bool_t circp, 
    bool_t diagonal,
    double skipProb
  );
  /* Generates a random unitary monotonic pairing {p},
    starting with a rung in the set {{0..nx-1}{0..ny-1}}. The
    parameters {nx} and {ny} must be positive.
    
    If {circx} is FALSE, the X coordinates or all rungs will be
    restricted to {0..nx-1}; if {circy} is FALSE, the Y coordinates
    are restricted to {0..ny-1}.  
    
    The pairing will cover approximately {len} bases on each sequence
    (it may vary somewhat, depending on other constraints).
    If either sequence is open, {len} must not exceed the
    corresponding sample count ({na} or {nb}).  In any case,
    the number of rungs is chosen by the procedure, normally
    between {len} and {2*len-1}.
    
    The pairing will be circular if and only if {circp} is TRUE. In
    that case, at least one of {circx} and {circy} must be TRUE. If
    {circx} is TRUE, the period {PX} will be some multiple of {nx}
    (possibly 0), otherwise it will be 0. Ditto for {circy,PY,ny}.
    
    If {diagonal} is TRUE, the endpoints are chosen near
    the diagonal of the rectangle {{0..nx-1}{0..ny-1}}. 
    The parameter {skipProb} specifies the approximate ratio
    of axial steps (`skips') to diagonal steps (`hops'). */

/* PAIRING I/O */

void dm_pairing_write(FILE *wr, int nx, int ny, dm_pairing_t *p, int ixsize);
  /* Writes the pairing {*p} to the file {wr}.
    
    The format is the number of pairs {ng} (which must be positive),
    the initial rung {ix,iy}, the coordinate period vector {PX,PY} (or
    0,0 if the pairing is open), the final rung {fx,fy}, and the rungs
    {p[0..ng-1]}. The integers are separated by whitespace.
    
    A perfect pairing is encoded by a single '*' character. Other
    pairings are encoded by an initial '|' character, then a sequence
    of {ng-1} steps, as described under {dm_pairing_step_write}. */

dm_pairing_t *dm_pairing_read(FILE *rd);
  /* Reads from the file {rd} a pairing {p} betqeen two sequences
    {x} and {y}, in the format produced by {dm_pairing_write}. */

/* PLOTTING */

void dm_pairing_paint(dm_pairing_t *p, int nx, int ny, float v, float_image_t *img, int scale);
  /* Paints the pairing {p}, reduced by {1/scale}, on
    image {img} with color {v}.
    
    More precisely, each rung {ix,iy} of {p} is reduced modulo the
    parameters {nx,ny}, to the rectangle {{0..nx-1}{0..ny-1}}, and
    the value {v} is added to pixel in column {ix/scale} and row
    {iy/scale} of {img}. */

#endif
