#ifndef dm_psplot_H
#define dm_psplot_H

/* Postscript plots of graphs and such. */
/* Last edited on 2006-12-30 20:43:18 by stolfi */

#define dm_psplot_H_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#include <dm_basic.h>

#include <vec.h>
#include <pswr.h>

typedef struct dm_psplot_t dm_psplot_t;
  /* A handle to a Postscript graph-plotting stream. */

#define dm_PT_PER_MM (72.0/25.4)
  /* Postscript points (pt) per millimeter (mm). */ 

#define dm_EPS_MARGIN_MM (1.0)
  /* EPS figure margin width (in mm). */
 
dm_psplot_t *dm_psplot_new
  ( FILE *wr,
    char *prefix,
    double hSize, 
    double vSize,
    double mrg
  );
  /* Writes to file {wr} the preamble of an Encapsulated Postscript
    figure of width {hSize} and height {vSize}, plus an extra margin
    of width {mrg} all around.
    
    The plottable area will be the rectangle {[0_hSize][0_vSize]]
    in device coordinates.  The client and device reference rectangles
    will be initialized to that area.
    
    The device All dimensions are in mm. Returns a
    {dm_psplot_t} object that contains the {wr} handle and some
    extra state. */

dm_psplot_t *dm_psplot_new_named
  ( char *name,
    char *tag,
    double hSize, 
    double vSize,
    double mrg
  );
  /* Same as {dm_psplot_init}, but writes to a new disk file called "{name}{tag}.eps". */

dm_psplot_t *dm_psplot_new_graph
  ( FILE *wr,
    char *prefix,
    double hGraphSize, 
    double vGraphSize,
    bool_t scaleL, bool_t titleL,
    bool_t scaleR, bool_t titleR,
    bool_t scaleB, bool_t titleB,
    bool_t scaleT, bool_t titleT,
    double fontSize,
    int maxLabChars,
    double mrg
  );
  /* Writes to file {wr} the preamble of an Encapsulated Postscript
    figure that is to contain a graph of width {hGraphSize} and height
    {vGraphSize}. Returns a {dm_psplot_t} object that
    contains the {wr} handle and some extra state. 
    
    The plottable area will include space for the graph, plus space
    around it for graph scales at the left, right, bottom or top (as
    specified by the flags {scaleL,scaleR,scaleB,scaleT} and/or titles
    (as specified by the flags {titleL,titleR,titleB,titleT}. The
    figure will have an extra margin of {mrg} mm all around the
    plottable area.  The width of the bands is estimated 
    from the {fontSize} parameter (the label font's nominal
    size, in pt) and {maxLabChars} (the max number of characters 
    in any label of the scale). 
    
    The client and device reference rectangles will be initialized
    to the graph area only (excluding the scale and title areas).
    
    All dimensions are in mm. */
  
void dm_psplot_close(dm_psplot_t *dp);
  /* Terminates any pictures that have been written to {dp}, 
    flushes and closes any open files. Then frees all internal
    storage used by {dp} (including the associated {PSStream} object). */

PSStream *dm_psplot_get_ps_stream(dm_psplot_t *dp);
  /* Returns the {PSStream} object associated with the {dm_psplot_t}
    object {dp}.  The {pswr.h} routines can be used to draw on that
    stream. For those routines, the client coordinates are
    measured (in mm) from the lower left corner of the usable
    plotting area. */

void dm_psplot_get_plot_size(dm_psplot_t *dp, double *hSize, double *vSize);
  /* Stores in {*hSize} and {*vSize} the dimensions (in mm) of the usableplot
    area, as specified to {dm_psplot_init}, excluding the margin. */
   
/* COORDINATE MAPPING
  
  Many plotting routines in this interface take their arguments in
  /client coordinates/ {x,y}. A {dm_psplot_t} object defines
  a mapping of those client coordinates to /device
  coordinates/ {h,v}, measured in millimeters from the lower left
  corner of the plot area. */
  
double dm_psplot_map_x(dm_psplot_t *dp, double x);
double dm_psplot_map_y(dm_psplot_t *dp, double y);
double dm_psplot_map_coord(dm_psplot_t *dp, PSAxis axis, double coord);
void dm_psplot_map_coords(dm_psplot_t *dp, double x, double y, double *h, double *v);
  /* These procedures map client coordinates {x,y} to device
     coordinates {h,v}, as defined by {dp}. */

double dm_psplot_unmap_h(dm_psplot_t *dp, double h);
double dm_psplot_unmap_v(dm_psplot_t *dp, double v);
double dm_psplot_unmap_coord(dm_psplot_t *dp, PSAxis axis, double coord);
void dm_psplot_unmap_coords(dm_psplot_t *dp, double h, double v, double *x, double *y);
  /* These procedures map device coordinates {h,v} to client
     coordinates {x,y}, as defined by {dp}. */
  
/* SETTING THE COORDINATE MAPPINGS

  The client-to-device mapping of a {dm_psplot_t} object is defined by
  two rectangles stored in the object: the /device reference window/
  {DRW = [hMin_hMax][vMin_vMax]} and the /client reference window/
  {CRW = [xMin_xMax][yMin_yMax]}.
  
  The client-to-device mapping is such that the {CRW} is mapped to the
  {DRW}. More precisely, {x=xMin} is mapped to {h=hMin}, {x=xMax} to
  {h=hMax}. So, if {hMin > hMax} (or {xMin > xMax}) the client {x}
  axis is directed from right to left. The same holds for the {y} and
  {v} coordinates.
  
  Changing either of these rectangles affects the way that client
  coordinates are interpreted by subsequent function calls.

  The rectangles {CRW} and {DRW} are not used for clipping.
  Regardless of them, the usable plotting area is always
  the rectangle {[0_hSize][0_vSize]} in device coordinates. */

void dm_psplot_set_device_ref_window
  ( dm_psplot_t *dp,
    double hMin, double hMax, 
    double vMin, double vMax
  );
  /* Sets the device reference window to the specified
    rectangle (in mm from the lower left corner of the 
    plottable area).*/

void dm_psplot_set_client_ref_window
  ( dm_psplot_t *dp,
    double xMin, double xMax, 
    double yMin, double yMax
  );
  /* Sets the client reference window to the specified
    rectangle (in client coordinates).*/

void dm_psplot_shrink_device_ref_window
  ( dm_psplot_t *dp,
    double lMrg, double rMrg, 
    double bMrg, double tMrg
  );
  /* Displaces the device reference window boundary INWARDS by the
    specified amounts on the left, right, bottom, and top sides,
    respectively. */

void dm_psplot_expand_client_ref_window
  ( dm_psplot_t *dp,
    double lMrg, double rMrg, 
    double bMrg, double tMrg
  );
  /* Displaces the client reference window boundary OUTWARDS by the
    specified amounts on the left, right, bottom, and top sides,
    respectively. */

/* DRAWING TOOLS 

  Unless specified otherwise, all procedures below take client
  coordinates. */

void dm_psplot_draw_segment(dm_psplot_t *dp, double xa, double ya, double xb, double yb);
  /* Draws a line segment from {(xa,ya)} to {(xb,yb)} with the current pen
    and draw color. */

void dm_psplot_draw_ref_axis(dm_psplot_t *dp, PSAxis axis);
  /* Draw the {x} or {y} coordinate axes in dark gray.
    The axis will span the current reference rectangle 
    (NOT the current plottable area). */

void dm_psplot_draw_tick
  ( dm_psplot_t *dp, 
    PSAxis axis, 
    double xt, 
    double yt,
    double tickSize,
    double tickAlign,
    char *label,
    double labAlign
  );
  /* Draws a tick mark, perpendicular to the given {axis}, at the
    nominal position {xt,yt}, with the current pen and draw color.
    The parameters {tickSize,tickAlign} are
    passed to {pswr_tick}.
    
    If the string {label} is not NULL, its is written using the
    current font and text color. The label is centered on the tick
    position along the given {axis}, and aligned as specified by
    {labAlign}, relative to the tick, along the other axis. Thus, for
    example, {labAlign=1.2} would be appropriate for a tick along the
    left or bottom edges of the plot area, and {labAlign=-0.2} would
    do for ticks along the top and right edges. */

void dm_psplot_draw_scale
  ( dm_psplot_t *dp, 
    PSAxis axis, 
    double pos,
    double tickSize,
    double tickAlign,
    double tickMinDist, 
    double tickMinStep, 
    char *fmt,
    double labAlign,
    double labMinDist,
    double labMinStep
  );
  /* Draws the tick marks for the specified {axis}, in black. 
    
    The ideal axis of the ticks (the /tickline/) will be at coordinate
    {pos} on the other axis. The set of tick marks will span the
    reference rectangle. The spacing of the ticks is chosen by the
    procedure, with the constraint that the spacing between tick marks
    (labeled or unlabeled) must be at least {tickMinDist} in millimeters, and
    {tickMinStep} in client coordinates. (At least one of these
    constraints must be positive.)
    
    If the parameter {fmt} is not NULL, a subset of the ticks will be
    labeled with the corresponding client coordinates, in the format
    {fmt} (as per {printf}). Labeled ticks are longer than unlabeled
    ones ({1.5*tickSize}, in mm). The spacing between labeled ticks
    will be at least {labMinDist} in millimeters, and {labMinStep}
    in client coordinates.  
    
    The parameters {tickSize,tickAlign,labAlign} are explained
    under {dm_psplot_draw_tick}. */

void dm_psplot_choose_tick_coords
  ( dm_psplot_t *dp, 
    PSAxis axis, 
    double cMin, 
    double cMax, 
    double minDist, 
    double minStep,
    double *zMinP,
    double *zMaxP,
    double *zStepP
  );
  /* Chooses client coordinates for minor ticks within the range
    {[cMin_cMax]} of device coordinates.
    
    Specifically, the procedure chooses a nice round increment
    {zStep} in client coordinates between two consecutive ticks, and
    computes the client coordinates {zMin,zMax}, multiples of
    {zStep}, of the first and last tick that fit in that range.
    
    The parameter {minDist}, if positive, and specifies the minimum
    distance (in mm) between ticks. The parameter {minStep}, if
    positive, specifies the minimum client coordinate increment
    between ticks. At least one of {minDist,minStep} must be
    positive.
    
    The results are returned in {*zMinP,*zMaxP,*zStepP}.
    If there are no nice coordinate values in the specified range,
    returns {*zStepP=0}, {*zMinP>*zMaxP}. */ 

void dm_choose_label_coords
  ( dm_psplot_t *dp,
    PSAxis axis,
    double ztMin,
    double ztMax, 
    double ztStep,
    double minDist,
    double minStep,
    int *labPerP,
    int *labSkpP
  );
  /* Chooses client coordinates for major (labeled) ticks,
    given the spacing {ztStep} (in client coordinates) between
    minor ticks, and the client coordinates {ztMin,ztMax} of the first and last
    minor ticks.

    Specifically, the procedure chooses the number {labPer} (a
    positive integer) of ordinary tick steps for each major tick step,
    so that the label increment {labPer*ztStep} is a nice round value.
    It also computes the number {lapSkp} of minor tick steps between
    {ztMin} and the first major tick after it.
    
    If {minDist} is positive, it specifies the minimum spacing (in mm)
    between labeled ticks. If {minStep} is positive, it specifies the
    minimum increment (in client coordinates) between consecutive
    labels.
    
    The results are returned in {*labPerP,*lasbSkpP}.  If the constraints
    imply that there are no major ticks in the range {ztMin,ztMax},
    returns {*labPerP=0}. */ 

void dm_psplot_draw_ref_frame(dm_psplot_t *dp);
  /* Draw a thin black frame around current reference rectangle
   (NOT the current plottable area). */

void dm_psplot_draw_plot_frame(dm_psplot_t *dp);
  /* Draw a thin black frame around current plottable area. */

void dm_psplot_draw_y_polyline
  ( dm_psplot_t *dp,
    double xMin, 
    double xMax,
    double y[],
    int n
  );
  /* Draws a polygonal line with {n} vertices and {n-1} line segments.
    The vertex X coordinates are univormly spaced in {[xMin_xMax]}.
    The Y coordinates are {y[0..n-1]}. */

void dm_psplot_draw_y_dots
  ( dm_psplot_t *dp,
    double xMin, 
    double xMax,
    double y[],
    int n,
    double rad,
    bool_t fill,
    bool_t draw
  );
  /* Draws {n} dots of radius {rad}. The X coordinates of the
    dot centers are uniformly spaced in {[xMin_xMax]}. The Y
    coordinates are {y[0..n-1]}.
    
    The radius {rad} is in mm, while the centers are in client
    coordinates. The parameters {fill} and {draw} are as in
    {pswr_dot}. */

#endif
