/* See dm_refine.h */
/* Last edited on 2007-01-04 01:45:45 by stolfi */

#define dm_refine_C_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)" \
  " and the Federal Fluminense University (UFF)"

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_rung.h>
#include <dm_pairing.h>
#include <dm_cand.h>
#include <dm_cand_list.h>
#include <dm_refine.h>
#include <dm_image.h>

#include <affirm.h>
#include <filefmt.h>
#include <float_image.h>

#include <assert.h>
#include <stdio.h>
#include <math.h>

/* INTERNAL PROTOTYPES */

void dm_refine_cand_fill_tableau
  ( dm_cand_t *cdold, 
    dm_seq_t *ap, 
    dm_seq_t *bp, 
    int delta,
    int kappa,
    bool_t shrink,
    dm_step_score_proc_t *step_score, 
    dm_rung_t *goptp,    /* OUT: last rung of optimum pairing. */
    double *voptp,       /* OUT: score of optimum pairing. */
    dm_dyn_tableau_t *tb /* WORK: dynamic programming tableau. */
  );
  /* Runs the incremental dynamic programming algorithm to compute
    an optimum pairing of the sequences {ap} and {bp} that 
    is a refinement of the pairing {cdold->pv}.  
    
    The new path is constrained to use only /valid/ rungs. The set of
    valid rungs is defined as the subset of the integer XY grid which
    is the union of squares of side {2*delta+1} centered on rungs of
    the old pairing, united with squares of side {kappa+1} whose
    maximum (resp. minimum) corner are the first (resp. last) rung of
    that pairing.
    
    The procedure fills the dynamic programming tableau{*tb}, which is
    expanded as needed. It also returns the last rung of the optimum
    pairing in {*goptp}, and its score in {*voptp}. The optimum
    pairing can be recovered by following the {prev} links in the
    tableau, starting with rung {*goptp}. 
    
    If {shrink} is TRUE, allows the optimum path to start and end 
    anywhere within the valid region.  If {shrink} is false, 
    the path must begin and end near the ends of the original
    pairing, as allowed by {delta} and {kappa}. */

void dm_refine_compute_r_range
  ( dm_pairing_t *p, 
    int delta, 
    int kappa, 
    int *minrp, 
    int *maxrp
  );
  /* Computes the minimum and maximum R-coordinates {*minrp,*maxrp} of any valid rung,
    given the old pairing {p} and the parameters {delta,kappa}. */
  
void dm_refine_compute_s_range
  ( dm_pairing_t *p,
    int r,
    int *kp,
    int delta,
    int kappa,
    int *minsp,
    int *maxsp
  );
  /* Computes the minimum and maximum S-coordinates {*minsp,*maxsp} of any valid rung,
    with R-coordinate {r}, given the old pairing {p} and the parameters {delta,kappa}. */

/* IMPLEMENTATIONS */

dm_cand_vec_t dm_refine_cand_vec
  ( dm_cand_vec_t *cdvold,
    dm_seq_t *ap, 
    dm_seq_t *bp,
    int delta,
    int kappa, 
    bool_t shrink,
    dm_step_score_proc_t *step_score,
    dm_dyn_tableau_t *tb, 
    int maxCands,
    double frac
  )
  { dm_cand_vec_t cdvnew = dm_cand_vec_new(cdvold->nel);
    int ic;
    int ncnew = 0; /* Number of candidates that were kept. */
    for (ic = 0; ic < cdvold->nel; ic++)
      { dm_cand_t cdnew = dm_refine_cand
          ( &(cdvold->el[ic]), ap, bp, delta, kappa, shrink, step_score, tb );
        dm_cand_list_insert(&cdvnew, &ncnew, maxCands, &cdnew, frac);
      }
    dm_cand_vec_trim(&cdvnew, ncnew);
    return cdvnew;
  }
    
dm_cand_t dm_refine_cand
  ( dm_cand_t *cdold, 
    dm_seq_t *ap, 
    dm_seq_t *bp, 
    int delta,
    int kappa,
    bool_t shrink,
    dm_step_score_proc_t *step_score, 
    dm_dyn_tableau_t *tb
  )
  { fprintf(stderr, "  ----------------------------------------------------------\n");
    fprintf(stderr, "    refining candidate\n");
    fprintf(stderr, "    old: ");
    dm_cand_write(stderr, cdold, 5, 10, 6);
    fprintf(stderr, "\n");
    
    /* We can't handle circular pairings yet: */
    bool_t circp = dm_pairing_is_circular(cdold->pr);
    demand((! circp), "circular pairings not supported yet");
    
    dm_rung_t gopt;
    double vopt;
    dm_refine_cand_fill_tableau
      ( cdold, ap, bp, delta, kappa, shrink, step_score, &gopt, &vopt, tb );

    /* Get number of rungs {nr} in optimum pairing: */
    int nr = 0;
    dm_rung_t g = gopt;
    while (! dm_rung_is_none(&g)) 
      { nr++; 
        dm_dyn_entry_t *ep = dm_dyn_tableau_get_entry_address(tb, g);
        g = (ep == NULL ? dm_rung_none : ep->prev);
      }
    /* Now collect the optimum pairing: */
    dm_rung_vec_t gvnew = dm_rung_vec_new(nr);
    g = gopt; int ip = nr;
    while (! dm_rung_is_none(&g)) 
      { /* fprintf(stderr, "      ip = %3d g = (%5d,%5d)\n", ip, g.c[0], g.c[1]); */
        ip--; gvnew.el[ip] = g; 
        dm_dyn_entry_t *ep = dm_dyn_tableau_get_entry_address(tb, g);
        g = (ep == NULL ? dm_rung_none : ep->prev);
      }

    /* Create candidate: */
    dm_pairing_t *pr = dm_pairing_from_rung_vec(&gvnew, circp);
    dm_cand_t cdnew = dm_cand_from_pairing(ap, bp, pr, vopt); 

    fprintf(stderr, "    new: ");
    dm_cand_write(stderr, &cdnew, 5, 10, 6);
    fprintf(stderr, "\n");

    fprintf(stderr, "  ----------------------------------------------------------\n");
    return cdnew;
  }
  
#define dm_refine_ADD_RUNG_SCORE FALSE
  /* If FALSE, the score of a pairing is the sum of its step scores only.
    If TRUE, it also includes the sum of its rung scores. */

void dm_refine_cand_fill_tableau
  ( dm_cand_t *cdold, 
    dm_seq_t *ap, 
    dm_seq_t *bp, 
    int delta,
    int kappa,
    bool_t shrink,
    dm_step_score_proc_t *step_score, 
    dm_rung_t *goptp,    /* OUT: last rung of optimum pairing. */
    double *voptp,       /* OUT: score of optimum pairing. */
    dm_dyn_tableau_t *tb /* WORK: dynamic programming tableau. */
  )
  { fprintf(stderr, "      filling matrices...\n");
    dm_pairing_t *prold = cdold->pr;
    
    /* We can't handle circular pairings yet: */
    demand((!dm_pairing_is_circular(prold)), "circular pairings not supported yet");
    
    int minr, maxr; /* Min and max R-coord of any valid rung. */
    dm_refine_compute_r_range(prold, delta, kappa, &minr, &maxr);

    /* Compute max differente {maxds} between two S-coordinates of valid rungs: */
    int maxds = 2*(kappa + 2*delta);
    /* Make sure that the tableau has all the required elements: */
    dm_dyn_tableau_resize(tb, minr, maxr, maxds);
    
    auto bool_t in_head(dm_rung_t g);
    auto bool_t in_tail(dm_rung_t g);
      /* TRUE if {g} is inside the `head' (resp. `tail') of the valid
        region, namely the square of side {kappa} whose upper right
        (resp. lower left) corner is the first (resp. last) rung of
        {prold}, widened by {delta} on all sides. */

    bool_t in_head(dm_rung_t g)
      { dm_rung_t gini = dm_pairing_get_rung(prold, 0);
        int dx = g.c[0] - gini.c[0];
        int dy = g.c[1] - gini.c[1];
        if ((dx > delta) || (dx < -kappa-delta)) { return FALSE; }
        if ((dy > delta) || (dy < -kappa-delta)) { return FALSE; }
        return TRUE;
      }
    
    bool_t in_tail(dm_rung_t g)
      { int nr = dm_pairing_num_rungs(prold);
        dm_rung_t gfin = dm_pairing_get_rung(prold, nr-1);
        int dx = g.c[0] - gfin.c[0];
        int dy = g.c[1] - gfin.c[1];
        if ((dx < -delta) || (dx > +kappa+delta)) { return FALSE; }
        if ((dy < -delta) || (dy > +kappa+delta)) { return FALSE; }
        return TRUE;
      }

    /* Fills the matrix within the band, in diagonal order.
      We process entries one diagonal at a time, in order of 
      increasing R-coordinate. */
    double vopt = -INF;   /* Score of the best path found so far. */
    dm_rung_t gopt = dm_rung_none;  /* Final rung {gopt} of that best path. */
    int r;
    int kp = 0; /* Approximate index in {p} of a rung with R-coord R; */
    for (r = minr; r <= maxr; r++)
      { /* Process tableau entries along the diagonal with R-coordinate {r}. */
        int mins, maxs; /* Range of S-coordinates for the R-coordinate {r}. */
        dm_refine_compute_s_range(prold, r, &kp, delta, kappa, &mins, &maxs);
        dm_dyn_tableau_set_s_range(tb, r, mins, maxs);

        /* Compute band elements in diagonal {r}: */
        int s;
        for (s = mins; s <= maxs; s += 2)
          { assert((r + s) % 2 == 0);
            int ix = (r + s) / 2;
            int iy = (r - s) / 2;
            /* fprintf(stderr, "      ix = %5d iy = %5d\n", ix, iy); */
            dm_rung_t g = (dm_rung_t){{ix, iy}};
            dm_dyn_entry_t *eg = dm_dyn_tableau_get_entry_address(tb, g);
            assert(eg != NULL);
            bool_t ix_inside = ((ap->circ) || ((ix >= 0) && (ix < ap->dv.nel)));
            bool_t iy_inside = ((bp->circ) || ((iy >= 0) && (iy < bp->dv.nel)));
            if ((! ix_inside) || (! iy_inside))
              { /* Rung {g} is not valid, set tableau entry to null values: */
                eg->score = -INF; eg->prev = dm_rung_none;
              }
            else
              { /* Compute the optimum path ending at rung {g}. */
                  
                double vsel = -INF;  /* Score of best path that ends at {g}. */
                dm_rung_t fsel;      /* Previous rung on that path. */
                
                auto void try_step(dm_rung_t f);
                  /* Computes the score of the best pairing that ends with rung
                    {f}, plus the cost of the step {f->g}.
                    If the step is valid, updates {fsel,vsel}. */
                    
                void try_step(dm_rung_t f)
                  { 
                    double vtof; /* Value of best path to {f}. */
                    if (dm_rung_is_none(&f))
                      { vtof = 0; }
                    else
                      { dm_dyn_entry_t *ef = dm_dyn_tableau_get_entry_address(tb, f);
                        vtof = (ef == NULL ? -INF : ef->score);
                      }
                    if (fabs(vtof) != INF)
                      { /* Rung {f} is valid: */
                        double vfg = step_score(f, g, ap, bp);
                        /* fprintf(stderr, "\n"); */
                        /* fprintf(stderr, "  step from (%d,%d)", f->c[0], f->c[1]); */
                        /* fprintf(stderr, " to (%d,%d) ", g.c[0], g.c[1]); */
                        /* fprintf(stderr, " score %9.6f + %9.6f = %9.6f", vtof, vfg, vf+vfg); */
                        /* fprintf(stderr, "\n"); */
                        double v = vtof + vfg;
                        if (v > vsel) { vsel = v; fsel = f; }
                      }
                  }
                
                /* Choose between the four optimum paths ending at
                  rung {g}. Namely, consider coming from rung {f00 =
                  dm_rung_none} (path starts at {g}) and from the
                  three admissible previous rungs {f01,f10,f11}.
                  In case of ties, consider first the diagonal step,
                  then the axial steps, then starting at {g}. */
                  
                try_step((dm_rung_t){{ ix-1, iy-1 }});
                try_step((dm_rung_t){{ ix,   iy-1 }});
                try_step((dm_rung_t){{ ix-1, iy   }});
                if (shrink || in_head(g)) { try_step(dm_rung_none); }

                /* Store optimum path and score in matrix entries for {g}: */
                eg->score = vsel; eg->prev = fsel;
                /* fprintf(stderr, "      prev[%5d,%5d]", g.c[0], g.c[1]); */
                /* fprintf(stderr, " = (%5d,%5d)\n", gsel.c[0], gsel.c[1]); */
                
                /* Update the best final rung {gopt} and score {vopt}: */
                if (shrink || in_tail(g))
                  { if (vsel > vopt) { vopt = vsel; gopt = g; } }
              }
          }
      }
    
    /* Return last rung and score of optimum pairing: */
    fprintf (stderr, "      vopt = %23.15f  gopt = (%5d,%5d)\n", vopt, gopt.c[0], gopt.c[1]);    
    (*goptp) = gopt; (*voptp) = vopt;
  }

void dm_refine_compute_r_range
  ( dm_pairing_t *p, 
    int delta, 
    int kappa, 
    int *minrp, 
    int *maxrp
  )
  {
    /* Get the first rung {rlo} of pairing: */
    int nr = dm_pairing_num_rungs(p);
    dm_rung_t rlo = dm_pairing_get_rung(p, 0);
    fprintf(stderr, "      lo rung = %5d %5d\n", rlo.c[0], rlo.c[1]);
    
    /* Get the last rung {rhi} of pairing: */
    dm_rung_t rhi = dm_pairing_get_rung(p, nr-1);
    fprintf(stderr, "      hi rung = %5d %5d\n", rhi.c[0], rhi.c[1]);
    
    /* Compute range {lor,hir} of R-coords of rungs in pairing {p}: */
    int lor = rlo.c[0] + rlo.c[1];
    int hir = rhi.c[0] + rhi.c[1];

    /* Compute minimum and maximum R-coord of any valid rung: */
    int minr = lor - 2*(delta + kappa);
    int maxr = hir + 2*(delta + kappa);

    /* Return results: */
    (*minrp) = minr;
    (*maxrp) = maxr;
  }
  
void dm_refine_compute_s_range
  ( dm_pairing_t *p,
    int r,
    int *kp,
    int delta,
    int kappa,
    int *minsp,
    int *maxsp
  )
  {
    /* Get num of rungs {nr} in given pairing: */
    int nr = dm_pairing_num_rungs(p);

    auto int rung_r(int k);
      /* Returns the R-coordinate ofrung {k} of pairing {p}. */
      
    auto int rung_s(int k);
      /* Returns the S-coordinate of rung {k} of pairing {p}. */
      
    int rung_r(int k)
      { dm_rung_t g = dm_pairing_get_rung(p, k);
        return g.c[0] + g.c[1];
      }
    
    int rung_s(int k)
      { dm_rung_t g = dm_pairing_get_rung(p, k);
        return g.c[0] - g.c[1];
      }

    /* Get the R- and S-coords {lor,los} of the first rung: */
    int lor = rung_r(0), los = rung_s(0);
    
    /* Get the R- and S-coords {hir,his} of the last rung */
    int hir = rung_r(nr-1), his = rung_s(nr-1);

    /* If {lor <= r <= hir}, find consecutive rung indices {kprev} and {knext}
      so that the step {p[kprev] -> p[knext]} crosses the sweepline;
      namely, such that {rung_r(kprev) <= r <= rung_r(knext)}. Note that
      {rung_r(k)} may increase by 1 or by 2 when {k} increases by 1.
      If {r < lor} or {r > hir} we set {knext} or {kprev}  to special values. */
    int kprev, knext;
    if (r < lor)
      { /* Sweepline lies before rung 0: */ 
        kprev = -1; knext = 0;
      }
    else if (r > hir)
      { /* Sweepline lies after rung {nr-1}: */
        kprev = nr-1; knext = nr;
      }
    else
      { /* Sweepline crosses the pairing {p}, must find {kprev,knext}. */
        /* Start the search for {kprev} with the given rung index {*kp}: */
        kprev = *kp; 
        if (kprev < 0) { kprev = 0; }
        if (kprev >= nr) { kprev = nr-1; }
        while ((kprev > 0) && (rung_r(kprev) > r)) { kprev--; }
        kprev++;
        while ((kprev < nr) && (rung_r(kprev) <= r)) { kprev++; }
        kprev--;
        assert((kprev >= 0) && (kprev < nr));
        
        /* Find {knext}: */
        knext = kprev;
        while ((knext < nr) && (rung_r(knext) < r)) { knext++; }
        assert(knext < nr);
      }
    /* Update approx index {*kp} for next call: */
    (*kp) = knext;

    /* Compute center {meds} and radius {rads} of the S-coord range for this R-coord: */
    int meds, rads;
    if (r <= lor - kappa)
      { meds = los; rads = 2*(kappa + delta) - (lor - r); }
    else if (r >= hir + kappa)
      { meds = his; rads = 2*(kappa + delta) - (r - hir); }
    else if (r < lor) 
      { meds = los; rads = lor - r + 2*delta; }
    else if (r > hir) 
      { meds = his; rads = r - hir + 2*delta; }
    else if (kprev == knext) 
      { meds = rung_s(kprev); rads = 2*delta; }
    else
      { meds = rung_s(kprev); rads = 2*delta - 1; }

    /* Compute range {mins..maxs} of the {s} coord, store in tableau: */
    int mins = meds - rads;
    int maxs = meds + rads;
    /* fprintf(stderr, "      r = %5d  mins = %5d maxs = %5d", r, mins, maxs); */
    /* fprintf(stderr, " (x,y)min = (%5d,%5d)", (r + mins)/2, (r - mins)/2); */
    /* fprintf(stderr, " (x,y)min = (%5d,%5d)", (r + maxs)/2, (r - maxs)/2); */
    /* fprintf(stderr, "\n"); */
    
    /* Return results: */
    (*minsp) = mins; (*maxsp) = maxs;
  }
