/* See dm_rung.h */
/* Last edited on 2007-01-04 01:58:02 by stolfi */

#define dm_rung_C_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)"

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_rung.h>
#include <dm_image.h>

#include <msm_basic.h>

#include <affirm.h>
#include <sign.h>
#include <fget.h>
#include <nget.h>
#include <jsrandom.h>

#include <assert.h>
#include <stdio.h>
#include <math.h>

bool_t dm_rung_equal(dm_rung_t *ga, dm_rung_t *gb)
  { return ((ga->c[0] == gb->c[0]) && (ga->c[1] == gb->c[1])); }

bool_t dm_rung_is_none(dm_rung_t *g)
  { return ((g->c[0] == dm_rung_none.c[0]) && (g->c[1] == dm_rung_none.c[1])); }

dm_rung_t dm_rung_throw(dm_rung_t *glo, dm_rung_t *ghi, bool_t diagonal)
  { int dx = ghi->c[0] - glo->c[0];
    int dy = ghi->c[1] - glo->c[1];
    dm_rung_t g = *glo;
    if (diagonal)
      { /* Pick a point on the diagonal: */
        double r = drandom();
        g.c[0] += (int)floor(r*(dx + 1));
        g.c[1] += (int)floor(r*(dy + 1));
      }
    else
      { /* Pick a point anywhere in the rectangle: */
        g.c[0] += abrandom(0, dx); 
        g.c[1] += abrandom(0, dy);
      }
    return g;
  }

dm_rung_t dm_rung_throw_middle(dm_rung_t *glo, dm_rung_t *ghi, double skipProb)
  { /* Number of axial steps on each side: */
    int DX = ghi->c[0] - glo->c[0];
    int DY = ghi->c[1] - glo->c[1];
    int nsmax = DX + DY; /* Max steps from {glo} to {ghi}. */
    int cs = (nsmax + abrandom(0,1))/2; /* Half the steps, rounded randomly */
    affirm((cs >= 0) && (cs <= nsmax), "bug");
    dm_rung_t g = *glo;
    if (DX > DY)
      { int dy = msm_choose(0, DY, skipProb);
        g.c[1] += dy;
        g.c[0] += cs - dy;
      }
    else
      { int dx = msm_choose(0, DX, skipProb);
        g.c[0] += dx;
        g.c[1] += cs - dx; 
      }
    /* Consistency check: */
    affirm(dm_rung_X(*glo) <= dm_rung_X(g), "bug");
    affirm(dm_rung_X(g) <= dm_rung_X(*ghi), "bug");
    affirm(dm_rung_Y(*glo) <= dm_rung_Y(g), "bug");
    affirm(dm_rung_Y(g) <= dm_rung_Y(*ghi), "bug");
    return g;
  }

dm_rung_vec_t dm_rung_vec_throw(dm_rung_t *gini, dm_rung_t *gfin, double skipProb)
  { /* Compute number {nc[j]} of datums covered by pairing in sequence {j}: */
    int ncx = gfin->c[0] - gini->c[0] + 1; /* Number of paired datums in seq 0. */
    int ncy = gfin->c[1] - gini->c[1] + 1; /* Number of paired datums in seq 1. */
    /* Allocate the rung vector: */
    dm_rung_vec_t gv = dm_rung_vec_new(ncx + ncy - 1);
    int ng = 0;
    /* Store initial rung: */
    gv.el[ng] = *gini; ng++;

    auto void append_rungs(dm_rung_t gi, dm_rung_t gf);
      /* Appends to {gv[0..ng-1]} a unitary monotonic list of
        random rungs, interpolating between {gi} (exclusive) and {gf}
        (inclusive). Assumes that the most recently
        appended rung {gv[ng-1]} is {gi}. */

    void append_rungs(dm_rung_t gi, dm_rung_t gf)
      { /* Number of samples covered on each side: */
        int nc0 = gf.c[0] - gi.c[0] + 1;
        int nc1 = gf.c[1] - gi.c[1] + 1;
        int nsmax = nc0 + nc1 - 2; /* Max steps from {gi} to {gf}. */
        if (nsmax <= 0)
          { /* Nothing to do. */ }
        else if (nsmax == 1)
          { /* Add the last rung: */
            dm_rung_vec_expand(&gv, ng);
            gv.el[ng] = gf; ng++;
          }
        else
          { /* Select a middle rung {gm} between {gi} and {gm}: */
            dm_rung_t gm = dm_rung_throw_middle(&gi, &gf, skipProb);
            /* The middle rung should be strictly between the first and last */
            affirm(dm_rung_strict_compare(&gi, &gm) < 0, "bug");
            affirm(dm_rung_strict_compare(&gm, &gf) < 0, "bug");
            /* Add rungs from {gi} (excl) up to {gm} (incl): */
            append_rungs(gi, gm);
            /* Add rungs from {gm} (excl) up to {gf} (incl): */
            append_rungs(gm, gf);
          }
      }

    append_rungs(*gini, *gfin);
    dm_rung_vec_trim(&gv, ng);
    return gv;
  }

bool_t dm_rung_step_is_monotonic(dm_rung_t *g, dm_rung_t *h, bool_t die)
  { int dx = h->c[0] - g->c[0];
    int dy = h->c[1] - g->c[1];
    if (dx < 0) { fail_test(die, "X step is not monotonic"); }
    if (dy < 0) { fail_test(die, "Y step is not monotonic"); }
    if ((dx == 0) && (dy == 0)) { fail_test(die, "null step"); }
    return TRUE;
  }

bool_t dm_rung_step_is_unitary_monotonic(dm_rung_t *g, dm_rung_t *h, bool_t die)
  { int dx = h->c[0] - g->c[0];
    int dy = h->c[1] - g->c[1];
    if ((dx < 0) || (dx > 1)) { fail_test(die, "X step is not 0 or 1"); }
    if ((dy < 0) || (dy > 1)) { fail_test(die, "Y step is not 0 or 1"); }
    if (dx + dy == 0) { fail_test(die, "null step"); }
    return TRUE;
  }

void dm_rung_bresenham(dm_rung_t g, dm_rung_t h, int *ngP, dm_rung_vec_t *gv)
  {
    int ng = (*ngP);
    /* Bresenham'h algorithm: */
    /* Compute total displacement {sgnd[j]*absd[j]} in each axis {j}: */
    int absd[2], sgnd[2];
    int j;
    for (j = 0; j <= 1; j++)
      { int d = h.c[j] - g.c[j];
        absd[j] = abs(d);
        sgnd[j] = (d == 0 ? 0 : (d < 0 ? -1 : +1));
      }
    /* Find axis {jmax} where all steps are nonzero: */
    int jmax = (absd[0] >= absd[1] ? 0 : 1);
    /* The other axis {jmin} may have zero steps */
    int jmin = 1 - jmax;
    /* Initialize Bresenham's remainder {rem}. */
    int rem = 0; /* Bresenham's remainder. */
    /* Interpolate rungs between {g} (incl) and {h} (excl): */
    while ((g.c[0] != h.c[0]) || (g.c[1] != h.c[1]))
      { /* Store the new rung {pnew}: */
        dm_rung_vec_expand(gv, ng);
        gv->el[ng] = g; ng++;
        /* Update {pnew} by Bresenham'h algorithm: */
        /* Advance {g.c[jmax]}, maybe {g.c[jmin]}: */
        g.c[jmax] += sgnd[jmax];
        rem += absd[jmin];
        if (2*rem >= absd[jmax])
          { g.c[jmin] += sgnd[jmin]; 
            rem -= absd[jmax];
          }
      }
    (*ngP) = ng;
  }

dm_rung_vec_t dm_rung_vec_map_coordinates(dm_rung_vec_t *gv, dm_rung_map_proc_t *map)
  { /* Get rung count {ng} and circularity {circp} of given rung_vec: */
    int ng = gv->nel;
    /* Allocate new rung vector: */
    dm_rung_vec_t gvnew = dm_rung_vec_new(ng);
    /* Map all rungs: */
    int i;
    for (i = 0; i < gvnew.nel; i++)
      { gvnew.el[i] = map(gv->el[i]); }
    /* Make rung_vec from rung vector: */
    return gvnew;
  }

dm_rung_vec_t dm_rung_vec_map_to_finer
  ( dm_rung_vec_t *gv, 
    int nxold, 
    int nyold, 
    int nxnew, 
    int nynew, 
    int nwtb
  )
  { 
    auto dm_rung_t map_rung(dm_rung_t g);
    
    dm_rung_t map_rung(dm_rung_t g)
      { int ix = dm_seq_index_map_to_finer(g.c[0], nxold, nxnew, nwtb);
        int iy = dm_seq_index_map_to_finer(g.c[1], nyold, nynew, nwtb);
        return (dm_rung_t){{ix,iy}};
      }
      
    return dm_rung_vec_map_coordinates(gv, &map_rung);
  }

dm_rung_vec_t dm_rung_vec_map_to_coarser
  ( dm_rung_vec_t *gv, 
    int nxold, 
    int nyold, 
    int nxnew, 
    int nynew, 
    int nwtb
  )
  { 
    auto dm_rung_t map_rung(dm_rung_t g);
    
    dm_rung_t map_rung(dm_rung_t g)
      { int ix = dm_seq_index_map_to_coarser(g.c[0], nxold, nxnew, nwtb);
        int iy = dm_seq_index_map_to_coarser(g.c[1], nyold, nynew, nwtb);
        return (dm_rung_t){{ix,iy}};
      }
      
    return dm_rung_vec_map_coordinates(gv, &map_rung);
  }



sign_t dm_rung_lex_compare(dm_rung_t *ga, dm_rung_t *gb)
  { if (ga->c[0] < gb->c[0])
      { return NEG; }
    else if (ga->c[0] > gb->c[0])
      { return POS; }
    else if (ga->c[1] < gb->c[1])
      { return NEG; }
    else if (ga->c[1] > gb->c[1])
      { return POS; }
    else 
      { return ZER; }
  }

sign_t dm_rung_strict_compare(dm_rung_t *ga, dm_rung_t *gb)
  { if ((ga->c[0] == gb->c[0]) && (ga->c[1] == gb->c[1]))
      { return ZER; }
    else if ((ga->c[0] <= gb->c[0]) && (ga->c[1] <= gb->c[1]))
      { return NEG; }
    else if ((ga->c[0] >= gb->c[0]) && (ga->c[1] >= gb->c[1]))
      { return POS; }
    else 
      { return ZER; }
  }

void dm_rung_step_write(FILE *wr, dm_rung_t *g, dm_rung_t *h)
  {
    int dx = h->c[0] - g->c[0];
    int dy = h->c[1] - g->c[1];
    if ((dx == 0) || (dy == 0) || (dx == dy))
      { /* Single argument: */
        int d;
        char ch;
        if (dx == dy)
          { d = dx; ch = '|'; }
        else if (dx == 0)
          { d = dy; ch = '\\'; }
        else if (dy == 0)
          { d = dx; ch = '/'; }
        else
          { assert(FALSE); }
        if (d != 1) { fprintf(wr, "%d", d); }
        fprintf(wr, "%c", ch);
      }
    else
      { /* Two arguments, different, neither of them zero: */
        fprintf(wr, "%d,%d|", dx, dy);
      }
  }

dm_rung_t dm_rung_step_read(FILE *rd, dm_rung_t g)
  { int ch = fgetc(rd);
    demand(ch != EOF, "unexpected EOF");
    /* Read arguments, if any: */
    int d = INT_MAX; /* Single argument. */
    int dx = INT_MAX, dy = INT_MAX; /* Argumetn pair. */
    if (((ch >= '0') && (ch <= '9')) || (ch == '+') || (ch == '-'))
      { ungetc(ch, rd);
        int ok = fscanf(rd, "%d", &d); 
        demand(ok != -1, "unexpected EOF");
        demand(ok == 1, "format error");
        ch = fgetc(rd);
        demand(ch != EOF, "unexpected EOF");
        if (ch == ',')
          { dx = d; d = INT_MAX;
            ok = fscanf(rd, "%d", &dy); 
            demand(ok != -1, "unexpected EOF");
            demand(ok == 1, "format error");
          }
        else
          { ungetc(ch, rd); }
      }
    /* Read slash charater ('\\', '/', or '|'): */
    ch = fgetc(rd);
    demand(ch != EOF, "unexpected EOF");
    if (ch == '/')
      { demand((dx == INT_MAX) && (dy == INT_MAX), "invalid '/' step");
        g.c[0] += (d == INT_MAX ? 1 : d);
      }
    else if (ch == '\\')
      { demand((dx == INT_MAX) && (dy == INT_MAX), "invalid '\\' step");
        g.c[1] += (d == INT_MAX ? 1 : d);
      }
    else if (ch == '|')
      { if (d != INT_MAX)
          { assert((dx == INT_MAX) && (dy == INT_MAX));
            dx = dy = d;
          }
        g.c[0] += (dx == INT_MAX ? 1 : dx);
        g.c[1] += (dy == INT_MAX ? 1 : dy);
      }
    else 
      { demand(FALSE, "bad character"); }
    return g;
  }

