/* See dm_seq.h */
/* Last edited on 2007-01-04 01:48:36 by stolfi */

#define dm_seq_C_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)" \
  " and the Federal Fluminense University (UFF)"

#include <dm_seq.h>
#include <dm_nucleic.h>
#include <dm_basic.h>
#include <dm_sample.h>
#include <dm_datum.h>
#include <dm_image.h>
#include <dm_psplot.h>

#include <vec.h>
#include <nget.h>
#include <fget.h>
#include <filefmt.h>
#include <jsrandom.h>
#include <pswr.h>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>

/* IMPLEMENTATIONS */

int dm_seq_length(dm_seq_t *seqp)
  { return seqp->dv.nel; }

bool_t dm_seq_is_circular(dm_seq_t *seqp)
  { return seqp->circ; }

dm_seq_t dm_seq_new(int n)
  { dm_seq_t seq;
    seq.id = dm_seq_id_none;
    seq.name = NULL;
    seq.cmt = NULL;
    seq.circ = FALSE;
    int k;
    for (k = 0; k < dm_CHANNELS; k++) { seq.sfac.f[k] = 1.0; }
    seq.dv = dm_datum_vec_new(n);
    return seq;
  }

dm_seq_t dm_seq_from_datum_vec
  ( dm_seq_id_t id, 
    char *name, 
    char *cmt, 
    bool_t circular, 
    dm_datum_scale_t *sfac,
    dm_datum_vec_t dv
  )
  { dm_seq_t seq;
    seq.id = id;
    seq.name = name;
    seq.cmt = cmt;
    seq.circ = circular;
    int k;
    for (k = 0; k < dm_CHANNELS; k++) { seq.sfac.f[k] = sfac->f[k]; }
    seq.dv = dv;
    return seq;
  }

#define dm_seq_version "2006-12-17"
#define dm_seq_type_name "encoded_bio_seq"

void dm_seq_write(FILE *fsq, dm_seq_t *seqp)
  { 
    /* Write the file header: */
    filefmt_write_header(fsq, dm_seq_type_name, dm_seq_version);
    
    /* Write the comment, if any: */
    if (seqp->cmt != NULL) { filefmt_write_comment(fsq, seqp->cmt, '|'); }
    
    /* Write the header fields: */
    fprintf(fsq, "circular = %s\n", (seqp->circ ? "T" : "F"));
    
    fprintf(fsq, "scale =");
    int k; for (k = 0; k < dm_CHANNELS; k++)
      { fprintf(fsq, " %24.16e", seqp->sfac.f[k]); }
    fprintf(fsq, "\n");
    
    fprintf(fsq, "channels = %d\n", dm_CHANNELS);
    
    fprintf(fsq, "elements = %d\n", seqp->dv.nel);
    
    /* Write the datum vector: */
    int ib;
    for (ib = 0; ib < seqp->dv.nel; ib++)
      { dm_datum_t *dp = &(seqp->dv.el[ib]);
        int k;
        for (k = 0; k < dm_CHANNELS; k++)
          { fprintf(fsq, " %+4d", dp->c[k]); }
        fprintf(fsq, "\n");
      }

    /* Write the file footer: */
    filefmt_write_footer(fsq, dm_seq_type_name);
  }

void dm_seq_write_named(dm_seq_t *seqp, char *name, char *tag)
  { FILE *wr = dm_open_write(name, tag, ".eqs", TRUE);
    dm_seq_write(wr, seqp);
    fclose(wr);
  }

dm_seq_t dm_seq_read(FILE *fsq)
  { 
    /* Check and skip the file header: */
    filefmt_read_header(fsq, dm_seq_type_name, dm_seq_version);
    
    /* Skip comment lines: */
    char *cmt = filefmt_read_comment(fsq, '|');
    
    /* Read the header fields: */
    bool_t circular = nget_bool(fsq, "circular"); fget_eol(fsq);
    
    dm_datum_scale_t sfac;
    sfac.f[0] = nget_double(fsq, "scale");
    int k; for (k = 1; k < dm_CHANNELS; k++) { sfac.f[k] = fget_double(fsq); }
    fget_eol(fsq);
    
    int chns = nget_int(fsq, "channels"); fget_eol(fsq);
    demand(chns = dm_CHANNELS, "wrong number of channels");
    
    int nbas = nget_int(fsq, "elements"); fget_eol(fsq);
    
    /* Read the datum vector: */
    dm_datum_vec_t dv = dm_datum_vec_new(nbas);
    int ib;
    for (ib = 0; ib < nbas; ib++)
      { dm_datum_t d;
        int k;
        for (k = 0; k < dm_CHANNELS; k++)
          { int s = fget_int(fsq);
            demand((s >= -127) && (s <= +127), "bad sample value"); 
            d.c[k] = s;
          }
        fget_eol(fsq);
        dv.el[ib] = d;
      }
    
    /* Check and skip file footer: */
    filefmt_read_footer(fsq, dm_seq_type_name);
    
    /* Create the {dm_seq_t} record: */
    dm_seq_id_t id = dm_seq_id_none;
    char *name = NULL;
    dm_seq_t seq = dm_seq_from_datum_vec(id, name, cmt, circular, &sfac, dv);
    return seq;
  }

dm_datum_t *dm_seq_get_datum_address(dm_seq_t *seqp, int i)
  { int n = seqp->dv.nel;
    if (seqp->circ) { i = i % n; if (i < 0) { i += n; } }
    assert((i >= 0) && (i < n));
    dm_datum_t *dp = &(seqp->dv.el[i]);
    return dp;
  }
  
dm_datum_t dm_seq_get_datum(dm_seq_t *seqp, int i)
  { dm_datum_t *dp = dm_seq_get_datum_address(seqp, i);
    return *dp;
  }
  
void dm_seq_set_datum(dm_seq_t *seqp, int i, dm_datum_t d)
  { dm_datum_t *dp = dm_seq_get_datum_address(seqp, i);
    (*dp) = d;
  }

dm_sample_t *dm_seq_get_sample_address(dm_seq_t *seqp, int i, int k)
  { assert((k >= 0) && (k < dm_CHANNELS));
    dm_datum_t *dp = dm_seq_get_datum_address(seqp, i);
    return &(dp->c[k]);
  }
  
dm_sample_t dm_seq_get_sample(dm_seq_t *seqp, int i, int k)
  { dm_sample_t *sp = dm_seq_get_sample_address(seqp, i, k);
    return *sp;
  }
  
void dm_seq_set_sample(dm_seq_t *seqp, int i, int k, dm_sample_t s)
  { dm_sample_t *sp = dm_seq_get_sample_address(seqp, i, k);
    (*sp) = s;
  }
 
int dm_seq_filtered_size(int nold, bool_t circ, int nwtb)
  { return (nold - (circ ? 0 : nwtb - 1))/2;
  }

dm_seq_t dm_seq_filter(dm_seq_t *oldp, double_vec_t *wtbp, char *wcmt)
  { int nold = oldp->dv.nel;
    int nwtb = wtbp->nel;
    int nnew = dm_seq_filtered_size(nold, oldp->circ, nwtb);
    if (nnew < 0) { nnew = 0; }
    /* Allocate new datum vector: */
    dm_datum_vec_t new_dv = dm_datum_vec_new(nnew);
    /* Scale factors for new samples: */
    dm_datum_scale_t new_sfac;
    int k;
    if (nnew == 0)
      { /* Sequence is too short to be filtered. */
        /* Set {new_sfac} to something sensible, just in case: */
        for (k = 0; k < dm_CHANNELS; k++) { new_sfac.f[k] = oldp->sfac.f[k]/M_SQRT2; }
        /* Vector {new_dv} is empty. */
      }
    else
      { 
        if (! oldp->circ) { assert(2*(nnew-1) + (nwtb-1) < nold); }

        /* Process datum vector, one channel at a time; */
        double_vec_t old_v = double_vec_new(nold); /* Decoded samples of old sequence. */
        double_vec_t new_v = double_vec_new(nnew); /* Decoded samples of new sequence. */
        for (k = 0; k < dm_CHANNELS; k++)
          { /* Extract and decode the old samples {old_v}: */
            double old_fac = oldp->sfac.f[k];
            int i;
            for (i = 0; i < nold; i++)
              { dm_datum_t *dp = dm_seq_get_datum_address(oldp, i);
                dm_sample_t s = dp->c[k];
                old_v.el[i] = dm_sample_decode(s)*old_fac; 
              }

            /* Filter the old samples to produce the new samples {new_v}: */
            dm_double_vec_filter(&old_v, wtbp, &new_v);

            /* Compute the new scale factor {new_fac}: */
            double sum_sq = 0;
            for (i = 0; i < nnew; i++) { double v = new_v.el[i]; sum_sq += v*v; }
            new_sfac.f[k] = sqrt(sum_sq/nnew);
            /* fprintf(stderr, "  channel %d sum_sq = %14.8f  sfac = %12.8f\n", k, sum_sq, new_sfac.f[k]); */

            /* Encode the new samples and store them into the new datum vector: */
            for (i = 0; i < nnew; i++)
              { dm_sample_t s = dm_sample_encode(new_v.el[i]/new_sfac.f[k]);
                dm_datum_t *dp = &(new_dv.el[i]);
                dp->c[k] = s; 
              }
          }

      }
      
    /* Append the weight table comment to the sequence's comment string: */
    char *new_cmt = NULL;
    if (wcmt == NULL) { wcmt = "unnamed table"; }
    char *old_cmt = (oldp->cmt != NULL ? oldp->cmt : "");
    asprintf(&(new_cmt), "%sfiltered with %s (window width %d)\n", old_cmt, wcmt, nwtb);

    /* Pack the filtered datum vector as a {dm_seq_t}: */
    dm_seq_t new = dm_seq_from_datum_vec
      ( oldp->id, oldp->name, new_cmt, oldp->circ, &new_sfac, new_dv );
    return new;
  }

void dm_double_vec_filter(double_vec_t *oldp, double_vec_t *wtbp, double_vec_t *newp)
  { int nold = oldp->nel;
    int nwtb = wtbp->nel;
    int nnew = newp->nel;
    int i, j;
    for (i = 0; i < nnew; i++)
      { double sum_w = 0, sum_ws = 0;
        for(j = 0; j < nwtb; j++)
          { double w = wtbp->el[j]; 
            double s = oldp->el[(2*i+j) % nold];
            sum_w += w; 
            sum_ws += w*s; 
          }
        newp->el[i] = sum_ws/sum_w;
      }
  }

int dm_seq_index_map_affinely(int ixold, int nold, int nnew)
  { /* Reduce the index {ixold} (which may be negative) modulo {nold}: */
    int kxold = (ixold % nold + nold) % nold;
    /* Map affinely the range {0..nold} to {0..nnew}: */
    int kxnew = (int)floor(((double)kxold + 0.5)*((double)nnew)/((double)nold));
    affirm((kxnew >= 0) && (kxnew < nnew), "bug");
    /* Add the original number {q} of full turns: */
    int q = (ixold - kxold)/nold;
    int ixnew = kxnew + q*nnew;
    return ixnew;
  }

int dm_seq_index_map_to_finer(int ixold, int nold, int nnew, int nwtb)
  { /* Reduce the index {ixold} (which may be negative) modulo {nold}: */
    int kxold = (ixold % nold + nold) % nold;
    /* Reverse the index map of {dm_double_vec_filter}. */
    int kxnew = 2*kxold + (nwtb-1)/2;
    affirm((kxnew >= 0) && (kxnew < nnew), "bug");
    /* Add the original number {q} of full turns: */
    int q = (ixold - kxold)/nold;
    int ixnew = kxnew + q*nnew;
    return ixnew;
  }

int dm_seq_index_map_to_coarser(int ixold, int nold, int nnew, int nwtb)
  { /* Reduce the index {ixold} (which may be negative) modulo {nold}: */
    int kxold = (ixold % nold + nold) % nold;
    /* Apply the index map of {dm_double_vec_filter}: */
    int kxnew = (kxold - (nwtb-1)/2)/2;
    /* If {nold} is odd, we may have {kxnew==nnew} if {ixold == nold-1}: */
    if (kxnew == nnew) 
      { assert(nold % 2 == 1);
        assert(ixold == nold-1);
        kxnew = nnew-1;
      }
    affirm((kxnew >= 0) && (kxnew < nnew), "bug");
    /* Add the original number {q} of full turns: */
    int q = (ixold - kxold)/nold;
    int ixnew = kxnew + q*nnew;
    return ixnew;
  }

void dm_seq_throw_segment(int ns, bool_t circ, int len, int *ini, int *fin)
  { int ix = abrandom(0, ns - (circ ? 1 : len)); 
    int fx = ix + len;
    assert((ix >= 0) && (ix < ns));
    assert(ix <= fx); 
    if (! circ) { assert(fx < ns); }
    (*ini) = ix;
    (*fin) = fx;
  }

dm_seq_t dm_seq_from_nucleic_string(dm_seq_id_t id, char *name, char *cmt, char *bas, bool_t circ)
  { /* Convert nucleotide codes to numeric datums: */
    dm_datum_vec_t dv = dm_datum_vec_from_nucleic_string(bas);
    /* Compute standard deviation {rms} of signal for uniformly distributed bases: */
    double rms = dm_NUCLEIC_RAW_SCALE;
    dm_datum_scale_t sfac;
    int k; for (k = 0; k < dm_CHANNELS; k++) { sfac.f[k] = rms; }
    /* Pack it all as a {dm_seq_t}: */
    dm_seq_t seq = dm_seq_from_datum_vec(id, name, cmt, circ, &sfac, dv); 
    return seq;
  }
  
dm_seq_t dm_seq_read_from_nucleic_file(dm_seq_id_t id, char *name, char *fname, bool_t circ)
  { /* Read letter sequence from file: */
    char *bas = NULL;
    char *cmt = NULL;
    dm_nucleic_string_read_named(&bas, &cmt, fname, "");
    /* Pack it all up: */
    dm_seq_t seq = dm_seq_from_nucleic_string(id, name, cmt, bas, circ);
    return seq;
  }
  
void dm_seq_postscript_plot_named
  ( dm_seq_t *seqp, 
    double hSize, 
    double vSize,
    double fontSize,
    char *name,
    char *tag
  )
  { /* Join {name} and {tag} into a file name prefix: */
    char *prefix = NULL;
    asprintf(&prefix, "%s%s", name, tag);
    /* Add the extension ".eps" and open the file: */
    FILE *wr = dm_open_write(name, tag, ".eps", TRUE);
    /* Create the {dm_psplot_t} object stream from it: */
    dm_psplot_t *dp = dm_psplot_new_graph
      ( wr, prefix, 
        hSize, vSize, 
        /*scaleL*/  TRUE,  /*titleL*/ FALSE,
        /*scaleR*/  FALSE, /*titleR*/ FALSE,
        /*scaleB*/  TRUE,  /*titleB*/ FALSE,
        /*scaleT*/  FALSE, /*titleT*/ FALSE,
        fontSize,
        /*maxLabChars*/ 7,
        dm_EPS_MARGIN_MM
      );
    
    /* Plot the sequence: */
    dm_seq_postscript_plot(dp, seqp);
    /* Close and cleanup: */
    dm_psplot_close(dp);
  }
  
void dm_seq_postscript_plot(dm_psplot_t *dp, dm_seq_t *seqp)
  { 
    /* Get the {PSStream} handle for {pswr.h} routines: */
    PSStream *ps = dm_psplot_get_ps_stream(dp);
    
    /* Get sequence size and circularity: */
    int nd = dm_seq_length(seqp); /* Number of datums. */
    bool_t circ = dm_seq_is_circular(seqp);
    int ns = (circ ? nd : nd - 1); /* Number of steps. */
    double *sfac = seqp->sfac.f; /* Sample scaling factor per channel. */
    if (nd == 0) { return; }
    
    /* If circular, the polyline has {ns+2} segments, from {x = -0.25} to {x = ns+0.25}. */
    /* If not circular, the polyline has {ns} segments, from {x = 0} to {x = ns}. */

    /* Choose nominal Y range {[yMin_yMax]}: */
    int c;
    double sfMax = 1.0e-100; /* Max sample scale factor among channels. */
    for (c = 0; c < dm_CHANNELS; c++)
      { if (sfac[c] > sfMax) { sfMax = sfac[c]; } }
    double vMax = dm_sample_decode(dm_sample_VALID_MAX)*sfMax;
    double ySkosh = 0.0;
    double yMin = -vMax - ySkosh;
    double yMax = +vMax + ySkosh;
    
    /* Choose nominal X range {[xMin_xMax]}: */
    double xWrap = 0.25; /* Extra fraction of step to show in circular seqs. */
    double xSkosh = xWrap + 0.02*ns;
    double xMin = 0 - xSkosh;
    double xMax = ns + xSkosh;
    
    /* Set client reference window, with some skosh: */
    dm_psplot_set_client_ref_window(dp, xMin, xMax, yMin, yMax);
    
    /* Shrink device window to leave space for ticks and labels: */
    double yScaleWd = 12.0; /* Width of Y scale and ticks. */
    double xScaleHt =  5.0; /* Height of X scale and ticks. */
    dm_psplot_shrink_device_ref_window(dp, yScaleWd, 0.0, xScaleHt, 0.0);
    
    /* Decide whether to plot dots at individual samples: */
    double hStep = dm_psplot_map_x(dp, 1) - dm_psplot_map_x(dp, 0);
    bool_t show_dots = (hStep >= 1.5);  /* If spaced at least 1.5 mm */

    /* Colors for channels (all with brightness = 0.30): */
    assert(dm_CHANNELS == 3);
    double R[3] = { 1.00, 0.00, 0.00 };
    double G[3] = { 0.00, 0.50, 0.33 };
    double B[3] = { 0.00, 0.00, 1.00 };
    
    /* Draw axes, ticks, labels, etc: */
    char *font = "Times-Roman";
    double fontSz = 12.0; /* In Postscript points. */
    double tickSz = 1.0;  /* In mm. */
    pswr_set_label_font(ps, font, fontSz);
    dm_psplot_draw_ref_axis(dp, HOR);
    dm_psplot_draw_scale(dp, HOR, yMin, tickSz,1.0, 5.0,1.0, "%.0f",   1.2, 12.0,1.0);
    dm_psplot_draw_scale(dp, VER, xMin, tickSz,1.0, 3.0,0.0, "%+.2f",  1.2,  4.0,0.0);

    for (c = 0; c < dm_CHANNELS; c++)
      { /* Plot channel {c}. */
        /* Set the pen color: */
        pswr_set_pen(ps, R[c], G[c], B[c], 0.25, 0.0, 0.0);
        /* Plot channel {c} of the sequence. */
        
        auto double yval(double x);
          /* Computes the client ordinate {y} for client abscissa {x},
            by linear interpolation between the decoded samples
            of channel {c} of sequence {seqp}. */
          
        double yval(double x)
          { int k = (int)floor(x);
            dm_sample_t s0 = dm_seq_get_sample(seqp, k, c);
            double y0 = dm_sample_decode(s0)*sfac[c];
            double s = x - (double)k;
            if (s == 0)
              { return y0; }
            else
              { dm_sample_t s1 = dm_seq_get_sample(seqp, k+1, c);
                double y1 = dm_sample_decode(s1)*sfac[c];
                double r = 1 - s;
                return r*y0 + s*y1;
              }
          }
        
        /* Compute first point {h0,v0} of polyline: */
        double h0, v0; /* Previously plotted point. */
        { double x0 = (circ ? -xWrap : 0);
          dm_psplot_map_coords(dp, x0, yval(x0), &h0, &v0);
        }
        /* Now plot all steps. */
        int i = (circ ? 0 : 1);
        while (i <= ns)
          { /* Plot step from sample {i-1} to sample {i}: */
            double x1 = (circ && (i >= ns) ? ns - xWrap : (double)i);
            double h1, v1;
            dm_psplot_map_coords(dp, x1, yval(x1), &h1, &v1);
            pswr_segment(ps, h0, v0, h1, v1);
            h0 = h1; v0 = v1;
            i++;
          }
        if (show_dots)
          { /* Draw dots at data points: */
            pswr_set_fill_color(ps, R[c], G[c], B[c]);
            for (i = 0; i < nd; i++)
              { dm_sample_t s = dm_seq_get_sample(seqp, i, c);
                double y = dm_sample_decode(s)*sfac[c];
                double h, v;
                dm_psplot_map_coords(dp, i, y, &h, &v);
                pswr_dot(ps, h, v, 0.5, TRUE, FALSE);
              }
          }
      }

    /* Draw a thin frame in black: */
    dm_psplot_draw_ref_frame(dp);
  }

