#define PROG_NAME "dm_test_discrim"
#define PROG_DESC "analysis of true/false candidate discrimination"
#define PROG_VERS "1.0"

/* Last edited on 2007-01-04 01:55:58 by stolfi */

#define dm_test_discrim_C_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#define PROG_HELP \
  PROG_NAME " \\\n" \
  "  -nBases {N_BASES} \\\n" \
  "  -mutProb {MUT_PROB} \\\n" \
  "  -insProb {INS_PROB} \\\n" \
  "  -maxLevel {MAX_LEVEL} \\\n" \
  "  -initFilter " dm_weight_args_HELP " \\\n" \
  "  -incrFilter " dm_weight_args_HELP " \\\n" \
  "  -scores " dm_score_args_HELP " \\\n" \
  "  [ -delta {DELTA} ] [ -kappa {KAPPA} ] \\\n" \
  "  -nCands {N_CANDS} \\\n" \
  "  {OUT_NAME}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  "  " PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  This program generates {2*N_CANDS} random candidate pairings" \
  " with same length {L} between two similar random" \
  " biosequences.  Half of the candidates are `true'" \
  " (associate homologous segments) and half are `false' (random). The" \
  " candidates are refined (once).  This procedure" \
  " is repeated for various filtering levels and" \
  " various candidate lengths {L}.  For each level, the mean scores" \
  " of the true and false candidates are plotted as a function" \
  " of candidate length {L}.\n" \
  "\n" \
  "  The sequences are derived from the same random sequence, with {N_BASES}" \
  " nucleotides, by random mutations, insertions and deletions.\n" \
  "\n" \
  "OUTPUT FILES\n" \
  "  All output files will have names starting with {OUT_NAME}.\n" \
  "\n" \
  "  The base-to-base distance matrix for the two sequences" \
  " is plotted as file \"{OUT_NAME}-eq.pgm\"." \
  "\n" \
  "OPTIONS\n" \
  "  -nBases {N_BASES}\n" \
  "    This mandatory argument specifies the" \
  " approximate length (number of nucleotide bases)" \
  " of the two test sequences.\n" \
  "\n" \
  "  -mutProb {MUT_PROB}\n" \
  "    This mandatory argument specifies the" \
  " nucleotide mutation probability used when" \
  " generating the two sequences.\n" \
  "\n" \
  "  -delProb {DEL_PROB}\n" \
  "    This mandatory argument specifies the" \
  " nucleotide deletion or insertion probability used" \
  " when generating the two sequences.\n" \
  "\n" \
  "  -maxLevel {MAX_LEVEL}\n" \
  "    This mandatory argument specifies the" \
  " maximum level of fouriering to be" \
  " applied.  Level 0 is the unfouriered" \
  " sequence, and each subsequent level is sampled at half" \
  " the frequency of the previous one.\n" \
  "\n" \
  "  -initFilter " dm_weight_args_HELP " \n" \
  "  -incrFilter " dm_weight_args_HELP " \n" \
  "    These mandatory arguments specify the" \
  " weights of the filter to be used at the" \
  " first filtering step and at subsequent" \
  " filtering steps.  " dm_weight_args_norm_sum_INFO "\n" \
  "\n" \
  "  -scores " dm_score_args_HELP "\n" \
  "    This mandatory argument specifies the" \
  " scoring of steps in a pairing.  " dm_score_args_INFO "\n" \
  "\n" \
  "  -delta {DELTA}\n" \
  "    This optional argument specifies the" \
  " amount of adjustment allowed for the X and Y coordinates of" \
  " internal rungs of each pairing.  The default is 3.\n" \
  "\n" \
  "  -kappa {DELTA}\n" \
  "    This optional argument specifies the" \
  " amount of X and Y extension allowed" \
  " at either end of each pairing.  The default is 6.\n" \
  "\n" \
  "  -nCands {N_CANDS}\n" \
  "    This mandatory argument specifies how many candidates" \
  " to generate.\n" \
  "\n" \
  argparser_help_info_HELP_INFO "\n" \
  "SEE ALSO\n" \
  "  dm_match(1)\n" \
  "\n" \
  "AUTHOR\n" \
  "  This program was created on 21/dec/2006 by J. Stolfi.\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " dm_test_discrim_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS  

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_nucleic.h>
#include <dm_rung.h>
#include <dm_score.h>
#include <dm_pairing.h>
#include <dm_cand.h>
#include <dm_refine.h>
#include <dm_dyn.h>
#include <dm_image.h>
#include <dm_weight.h>
#include <dm_multi.h>
#include <dm_test_tools.h>

#include <affirm.h>
#include <float_image.h>
#include <float_pnm_image.h>
#include <jspnm_image.h>
#include <argparser.h>

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

typedef struct dm_options_t 
  { /* Sequence generation parameters: */
    double nBases;       /* Expected number of bases. */
    double mutProb;      /* Mutation probability. */
    double delProb;      /* Insertion/deletion probability. */
    /* Candidate generation parameters: */
    int nCands;          /* Number of candidates to generate. */
    /* Filtering parameters: */
    int maxLevel;        /* Maximum filtering level. */
    double_vec_t w0;     /* Filter weights for first stage (level 0 to level 1). */
    double_vec_t w1;     /* Filter weights for subsequent stages. */
    /* Refinement parameters: */
    int delta;           /* Half-width of tableau around original pairing. */
    int kappa;           /* Extension of tableau beyond ends of pairing. */
    dm_score_args_t sc;  /* Scoring criterion. */
    /* Output parameters: */
    char *outName;       /* Output file name prefix (minus extensions). */
  } dm_options_t;
  
int main(int argc, char**argv);

dm_options_t *dm_get_options(int argc, char**argv);
  /* Parses the command line options, packs 
    them into a {dm_options_t} record. */

void dm_test_discrim_at_level
  ( dm_options_t *o, 
    int level, 
    dm_seq_t *xf, 
    dm_seq_t *yf,
    dm_rung_vec_t *gv
  );
  /* Run the discriminant analysis on sequences {xf,yf} which are filtered
    at the given {level}.  The rungs {gv[0..gv.nel-1]} should be the 
    index pairs of corresponding elements in {xf} and {yf}. 
    The rung list {gv} must be monotonic but need not be unitary. */

dm_cand_vec_t dm_test_discrim_throw_cands_false
  ( int nc, 
    int len, 
    dm_seq_t *x, 
    dm_seq_t *y, 
    double skipProb
  );
  /* Generates a list of {nc} random candidates for the two
    {dm_seq_t}s, each pairing about {len} elements on each sequence.
    Each step of each candidate will be a `skip' with probability
    {skipProb} */
  
dm_cand_vec_t dm_test_discrim_throw_cands_true
  ( int nc, 
    int len, 
    dm_seq_t *x, 
    dm_seq_t *y, 
    dm_rung_vec_t *gv
  );
  /* Generates a list of {nc} random candidates for the two
    {dm_seq_t}s, each pairing about {len} elements on each sequence.
    The candidates are cut out from the pairing {pr}. */
    
int main(int argc, char**argv)
  { 
    dm_options_t *o = dm_get_options(argc, argv);
    
    fprintf(stderr, "generating the ancestral sequence ...\n");
    char *borg = dm_nucleic_string_throw(o->nBases, TRUE, TRUE);
    
    fprintf(stderr, "writing out the ancestral sequence ...\n");
    dm_nucleic_string_write_named(borg, "Ancestral", o->outName, "-org");
    
    fprintf(stderr, "generating and writing the mutated copies ...\n");
    dm_seq_t x, y;  /* The two homologous sequences. */
    dm_rung_vec_t gv; /* The true pairs. */
    dm_test_tools_make_seq_pair
      ( borg, o->mutProb, o->delProb, FALSE, 0, "X", &x, 1, "Y", &y, &gv, o->outName );
    
    fprintf(stderr, "getting the weight tables ...\n");
    char *wname0 = dm_weight_make_descr(&(o->w0), "%6.4f");
    char *wname1 = dm_weight_make_descr(&(o->w1), "%6.4f");

    fprintf(stderr, "trying to filter up to level %d ...\n", o->maxLevel);
    dm_seq_t xf[o->maxLevel + 1];
    dm_multi_seq_filter(&x, o->maxLevel, &(o->w0), wname0, &(o->w1), wname1, xf);
    dm_seq_t yf[o->maxLevel + 1];
    dm_multi_seq_filter(&y, o->maxLevel, &(o->w0), wname0, &(o->w1), wname1, yf);
    dm_rung_vec_t gf[o->maxLevel + 1];
    dm_multi_rung_filter(&gv, o->maxLevel, gf);

    /* Figure out max useful level {maxlev}: */
    int maxlev = o->maxLevel;
    while 
      ( (maxlev >= 0) && 
        ( (dm_seq_length(&(xf[maxlev])) == 0) ||
          (dm_seq_length(&(yf[maxlev])) == 0)
        )
      ) { maxlev--; }
    fprintf(stderr, "maximum useful level is %d\n", maxlev);

    fprintf(stderr, "wrting and plotting the filtered sequences ...\n");
    dm_test_tools_write_and_plot_multi_seq(xf, maxlev, o->outName);
    dm_test_tools_write_and_plot_multi_seq(yf, maxlev, o->outName);

    fprintf(stderr, "running analyses ...\n");
    int level;
    for (level = 0; level <= maxlev; level++)
      { dm_test_discrim_at_level(o, level, &(xf[level]), &(yf[level])); }
  
    /* ...and we are done: */
    fprintf(stderr, "done.\n");
    return 0;
  }
      
void dm_test_discrim_at_level
  ( dm_options_t *o,
    int level, 
    dm_seq_t *xf, 
    dm_seq_t *yf,
    dm_rung_vec_t *gv
  )
  {
    int nc = o->nCands;

    /* Sequence lengths: */
    int nx = dm_seq_length(xf);
    int ny = dm_seq_length(yf);
    
    auto double step_score(dm_rung_t g, dm_rung_t h, dm_seq_t *xp, dm_seq_t *yp);

    double step_score(dm_rung_t g, dm_rung_t h, dm_seq_t *xp, dm_seq_t *yp)
      { return dm_score_step(&(o->sc), g, h, xp, yp); }

    fprintf(stderr, "plotting distance matrix for test seq pair ...\n");
    dm_pairing_seq_seq_image_write_named(xf, yf, step_score, o->outName, "-eq");
    
    int len;  /* Length of candidates. */
    int lenMax = (nx < ny ? nx : ny);
    int lenStep = lenMax / 20;
    int lenMin = lenStep; 
    for (len = lenMin; len <= lenMax; len += lenStep)
      { dm_cand_vec_t cdv[2];
        int cdkind; /* Kind of candidates: 0 = false, 1 = true. */
        for (cdkind = 0; cdkind < 2; cdkind++)
          { fprintf(stderr, "creating %d candidates of kind %d and length %d ...\n", nc, cdkind, len);
            dm_cand_vec_t *cdvp = &(cdv[cdkind]);
            if (cdkind == 0)
              { /* Generate some random candidates: */
                (*cdvp) = dm_test_discrim_throw_cands_false(nc, len, xf, yf, o->mutProb); 
              }
            else
              { /* Generate some true candidates: */
                (*cdvp) = dm_test_discrim_throw_cands_true(nc, len, xf, yf, gv); 
              }

            fprintf(stderr, "refining candidates ...\n");
            dm_dyn_tableau_t tb = dm_dyn_tableau_new(); /* Dynamic programming tableau. */
            cdv[cdkind] = dm_refine_cand_vec
              ( cdvp, xf, yf, 
                o->delta, o->kappa, FALSE,
                &step_score, &tb, 
                nc, -1.000
              );
            fprintf(stderr, "computing mean score ...\n");
            double sum_sc = 0;
            int i;
            for (i = 0; i < cdvp->nel; i++)
              { dm_cand_t *cd = &(cdvp->el[i]); 
                sum_sc += cd->score; 
              }
            double avg = sum_sc/cdvp->nel;
            fprintf(stderr, "average score for length %d = %10.4f\n", len, avg);
            /* !!! Should save scores for plotting. */
          }
      }
      
    fprintf(stderr, "writing scores ...\n");
    /* !!! TO FINISH */
  }
    
dm_cand_vec_t dm_test_discrim_throw_cands_false
  ( int nc, 
    int len, 
    dm_seq_t *x, 
    dm_seq_t *y, 
    double skipProb
  )
  { 
    /* Sequence lengths and circularities: */
    int nx = dm_seq_length(x);
    int ny = dm_seq_length(y);

    /* Allocate the candidate vector: */
    dm_cand_vec_t cdv = dm_cand_vec_new(nc);
    int i; 
    for (i = 0; i < nc; i++) 
      { cdv.el[i] = dm_cand_throw
          ( x->id, x->name, nx, FALSE, 
            y->id, y->name, ny, FALSE,
            len,
            /*circp*/ FALSE, 
            /*diagonal*/ FALSE,
            /*skipProb*/ skipProb
          );
      }
    return cdv;
  }

dm_cand_vec_t dm_test_discrim_throw_cands_true
  ( int nc, 
    int len, 
    dm_seq_t *x, 
    dm_seq_t *y, 
    dm_pairing_t *pr
  )
  { 
    /* Sequence lengths and circularities: */
    int nx = dm_seq_length(x);
    int ny = dm_seq_length(y);

    /* Allocate the candidate vector: */
    dm_cand_vec_t cdv = dm_cand_vec_new(nc);
    int i; 
    for (i = 0; i < nc; i++) 
      { /* Select the next piece of {gv}: */
        cdv.el[i] = dm_cand_throw_subcand(, len);
      }
    return cdv;
  }

dm_options_t *dm_get_options(int argc, char**argv)
  { 
    dm_options_t *o = (dm_options_t *)notnull(malloc(sizeof(dm_options_t)), "no mem");
    
    argparser_t *pp = argparser_new(stderr, argc, argv);
    argparser_set_help(pp, PROG_HELP);
    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);
    
    argparser_get_keyword(pp, "-nBases");
    o->nBases = argparser_get_next_int(pp, 1, INT_MAX);
    
    argparser_get_keyword(pp, "-mutProb");
    o->mutProb = argparser_get_next_double(pp, 0.0, 1.0);
    
    argparser_get_keyword(pp, "-delProb");
    o->delProb = argparser_get_next_double(pp, 0.0, 1.0 - o->mutProb);
    
    argparser_get_keyword(pp, "-initFilter");
    o->w0 = dm_weight_args_parse(pp, TRUE);

    argparser_get_keyword(pp, "-incrFilter");
    o->w1 = dm_weight_args_parse(pp, TRUE);

    argparser_get_keyword(pp, "-maxLevel");
    o->maxLevel = argparser_get_next_int(pp, 0, 20);

    argparser_get_keyword(pp, "-scores");
    o->sc = dm_score_args_parse(pp);
    
    if (argparser_keyword_present(pp, "-delta"))
      { o->delta = argparser_get_next_int(pp, 0, INT_MAX); }
    else
      { o->delta = 3; }
    
    if (argparser_keyword_present(pp, "-kappa"))
      { o->kappa = argparser_get_next_int(pp, 0, INT_MAX); }
    else
      { o->kappa = 6; }
    
    argparser_get_keyword(pp, "-nCands");
    o->nCands = argparser_get_next_int(pp, 0, INT_MAX);
    
    argparser_skip_parsed(pp);
    
    o->outName = argparser_get_next(pp);

    argparser_finish(pp);
    
    return o;
  }
