#define PROG_NAME "dm_test_scoring"
#define PROG_DESC "test of rung and step scoring functions"
#define PROG_VERS "1.0"

/* Last edited on 2007-01-04 00:12:13 by stolfi */

#define dm_test_scoring_C_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#define PROG_HELP \
  PROG_NAME " \\\n" \
  "  -scores " dm_score_args_HELP " \\\n" \
  "  [ -symmetry ] \\\n" \
  "  {OUT_NAME}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  "  " PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  This program prints the numerical score for various" \
  " rungs and steps between typical datums.\n" \
  "OUTPUT FILES\n" \
  "  All output files will have names starting with {OUT_NAME}.\n" \
  "\n" \
  "OPTIONS\n" \
  "  -scores " dm_score_args_HELP "\n" \
  "    This mandatory argument specifies the" \
  " scoring criterion to be used when evaluating" \
  " and comparing candidates.  " dm_score_args_INFO "\n" \
  "\n" \
  "  -symmetry\n" \
  "    If this optional argument is present, the" \
  " program will check whether the scoring functions" \
  " are symmetric under X-Y exchange.  If the" \
  " argument is omitted, the program" \
  " assumes the functions are symmetric, and skips the" \
  " testing of rungs and steps that differ only by X-Y" \
  " exchange.\n" \
  "\n" \
  argparser_help_info_HELP_INFO "\n" \
  "SEE ALSO\n" \
  "  dm_match(1)\n" \
  "\n" \
  "AUTHOR\n" \
  "  This program was created on 21/dec/2006 by J. Stolfi.\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " dm_test_scoring_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS  

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_nucleic.h>
#include <dm_rung.h>
#include <dm_score.h>

#include <bool.h>
#include <affirm.h>
#include <argparser.h>

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

typedef struct dm_options_t 
  { dm_score_args_t sc;   /* Scoring criterion. */
    bool_t symmetry;      /* If TRUE also checks X-Y symmetry. */
    /* Output parameters: */
    char *outName;     /* Output file name prefix (minus extensions). */
  } dm_options_t;
  
int main(int argc, char**argv);

dm_datum_t dm_datum_rescale(double alpha, dm_datum_t d) ;
  /* Multiplies all channels of datum {d} by {alpha}. Note that if {d}
   is not in {-1,0,+1}, the result will almost always include some
   quantization error. */

void dm_test_rung_scores(double mag, bool_t sym, dm_score_args_t *sc);
  /* Tests the single-rung scoring functions for many
    nucleotide pairs, with sample values scaled by {mag}.
    If {sym} is TRUE also checks whether the scores are
    symmetric under X-Y exchange. */

double dm_test_rung_score(double mag, dm_score_args_t *sc, char cx, char cy);
  /* Tests the single-rung scoring functions for the 
    nucleotide pair {cx,cy}, with sample values scaled by {mag}.
    Returns the step's score. */

void dm_test_step_scores(double mag, bool_t sym, dm_score_args_t *sc);
  /* Tests the step scoring functions for many 
    pairs of nucleotide pairs, with sample values scaled by
    {mag}.  If {sym} is TRUE also checks whether the scores are
    symmetric under X-Y exchange. */

double dm_test_step_score(double mag, dm_score_args_t *sc, char cgx, char cgy, char chx, char chy);
  /* Tests the step scoring functions for the pairing step between a
    rung {g} and a rung {h}, assuming {g} pairs nucleotides {cgx,cgy},
    and {h} pairs nucleotides {chx,chy}, with sample values scaled by
    {mag}.  Shows the scores for a single diagonal step and 
    two axial steps, in both orders.  Returns a weighted average
    of these three scores. */

dm_options_t *dm_get_options(int argc, char**argv);
  /* Parses the command line options, packs 
    them into a {dm_options_t} record. */

int main(int argc, char**argv)
  { 
    dm_options_t *o = dm_get_options(argc, argv);
    
    dm_test_rung_scores(1.0, o->symmetry, &(o->sc));
    dm_test_rung_scores(0.5, o->symmetry, &(o->sc));

    dm_test_step_scores(1.0, o->symmetry, &(o->sc));
    dm_test_step_scores(0.5, o->symmetry, &(o->sc));

    return 0;
  }
  
dm_datum_t dm_datum_rescale(double alpha, dm_datum_t d) 
  { int c;
    for (c = 0; c < dm_CHANNELS; c++) 
      { d.c[c] = dm_sample_encode(alpha*dm_sample_decode(d.c[c]));}
    return d;
  }

void dm_test_rung_scores(double mag, bool_t sym, dm_score_args_t *sc)
  { char *b = "ATCG"; /* The four nucleotides. */
    
    fprintf(stderr, "----------------------------------------------------------------------\n");
    fprintf(stderr, "Testing rung scores (mag = %8.6f) ...\n", mag);
    fprintf(stderr, "...\n");
    int x, y;
    /* Enumerate all rungs {(x,y)} {x <= y} : */
    for (x = 0; x < 4; x++)
      { for (y = x; y < 4; y++)
          { double scXY = dm_test_rung_score(mag, sc, b[x], b[y]);
            /* Scoring should be symmetric, but let's check: */
            if (x != y) 
              { double scYX = dm_test_rung_score(mag, sc, b[y], b[x]);
                affirm(scXY = scYX, "rung scoring is not X-Y symmetric");
              }
            fprintf(stderr, "\n");
          }
      }
    fprintf(stderr, "----------------------------------------------------------------------\n");
  }

double dm_test_rung_score(double mag, dm_score_args_t *sc, char cx, char cy)
  { dm_datum_t dx = dm_datum_rescale(mag, dm_datum_encoded_from_nucleic_char(cx));
    dm_datum_t dy = dm_datum_rescale(mag, dm_datum_encoded_from_nucleic_char(cy));
    double diffsq = dm_datum_diffsq(&dx, &dy);
    double score = dm_score_datum_rung(sc->eql, sc->dif, &dx, &dy);
    fprintf(stderr, "%c = ", cx);
    dm_datum_decoded_write(stderr, &dx, "( ", " ", " )");
    fprintf(stderr, "  ");
    fprintf(stderr, "%c = ", cy);
    dm_datum_decoded_write(stderr, &dy, "( ", " ", " )");
    fprintf(stderr, "  ");
    fprintf(stderr, " d2 = %8.6f ", diffsq);
    fprintf(stderr, " sc = %+9.6f", score),  
    fprintf(stderr, "\n");
    return score;
  }

void dm_test_step_scores(double mag, bool_t sym, dm_score_args_t *sc)
  { char *b = "ATCG"; /* The four nucleotides. */
    
    fprintf(stderr, "----------------------------------------------------------------------\n");
    fprintf(stderr, "Testing step scores (mag = %8.6f) ...\n", mag);
    fprintf(stderr, "...\n");
    int xg, yg, xh, yh;
    /* Enumerate all steps {(xg,yg) -> (xh,yh)} where {(xg,xh) <= (yg,yh)} : */
    for (xg = 0; xg < 4; xg++)
      { for (yg = xg; yg < 4; yg++)
          { for (xh = 0; xh < 4; xh++)
              { for (yh = (xg == yg ? xh : 0); yh < 4; yh++)
                  { double scXY = dm_test_step_score(mag, sc, b[xg], b[yg], b[xh], b[yh]);
                    /* Scoring should be symmetric, but let's check: */
                    if ((xg != yg) || (xh != yh))
                      { double scYX = dm_test_step_score(mag, sc, b[yg], b[xg], b[yh], b[xh]);
                        affirm(scXY = scYX, "step scoring is not X-Y symmetric");
                      }
                    fprintf(stderr, "\n");
                  }
              }
          }
      }
    fprintf(stderr, "----------------------------------------------------------------------\n");
  }

double dm_test_step_score(double mag, dm_score_args_t *sc, char cgx, char cgy, char chx, char chy)
  { dm_datum_t dgx = dm_datum_rescale(mag, dm_datum_encoded_from_nucleic_char(cgx));
    dm_datum_t dgy = dm_datum_rescale(mag, dm_datum_encoded_from_nucleic_char(cgy));
    dm_datum_t dhx = dm_datum_rescale(mag, dm_datum_encoded_from_nucleic_char(chx));
    dm_datum_t dhy = dm_datum_rescale(mag, dm_datum_encoded_from_nucleic_char(chy));
    fprintf(stderr, "%c %c = ", cgx, chx);
    dm_datum_decoded_write(stderr, &dgx, "( ", " ", " )");
    dm_datum_decoded_write(stderr, &dhx, "( ", " ", " )");
    fprintf(stderr, "\n");
    fprintf(stderr, "%c %c = ", cgy, chy);
    dm_datum_decoded_write(stderr, &dgy, "( ", " ", " )");
    dm_datum_decoded_write(stderr, &dhy, "( ", " ", " )");
    fprintf(stderr, "\n");
    double diffsq = dm_datum_step_diffsq(&dgx, &dgy, &dhx, &dhy);
    fprintf(stderr, "step d2 = %8.6f\n", diffsq);
    /* Increment penalties for diagonal step and two axial steps: */
    double s_diag = dm_score_step_increments(sc->skp, 1,1);
    double s_corn = 
      dm_score_step_increments(sc->skp, 0,1) + 
      dm_score_step_increments(sc->skp, 1,0);
    /* Score for a diagonal step: */
    double d_gg_hh = dm_score_datum_step(sc->eql, sc->dif, &dgx, &dgy, &dhx, &dhy);
    double score_diag = d_gg_hh + s_diag;
    fprintf
      ( stderr, 
        "score || =  %+9.6f + %+9.6f = %+9.6f\n", 
        d_gg_hh, s_diag, score_diag
      );  
    /* Score for a X-then-Y-step pair: */
    double d_gg_hg = dm_score_datum_step(sc->eql, sc->dif, &dgx, &dgy, &dhx, &dgy);
    double d_hg_hh = dm_score_datum_step(sc->eql, sc->dif, &dhx, &dgy, &dhx, &dhy);
    double score_X_then_Y = d_gg_hg + d_hg_hh + s_corn;
    fprintf
      ( stderr, 
        "score |/| = %+9.6f + %+9.6f + %+9.6f = %+9.6f + %+9.6f = %+9.6f\n", 
        d_gg_hg, d_hg_hh, s_corn, d_gg_hg + d_hg_hh, s_corn, 
        score_X_then_Y
      );  
    /* Score for a Y-then-X-step pair: */
    double d_gg_gh = dm_score_datum_step(sc->eql, sc->dif, &dgx, &dgy, &dgx, &dhy);
    double d_gh_hh = dm_score_datum_step(sc->eql, sc->dif, &dgx, &dhy, &dhx, &dhy);
    double score_Y_then_X = d_gg_gh + d_gh_hh + s_corn;
    fprintf
      ( stderr, 
        "score |\\| = %+9.6f + %+9.6f + %+9.6f = %+9.6f + %+9.6f = %+9.6f\n", 
        d_gg_gh, d_gh_hh, s_corn, d_gg_gh + d_gh_hh, s_corn, 
        score_Y_then_X
      );  
    fprintf(stderr, "\n");
    return (3*score_diag + 5*score_X_then_Y + 7*score_X_then_Y)/15;
  }
    
dm_options_t *dm_get_options(int argc, char**argv)
  { 
    dm_options_t *o = (dm_options_t *)notnull(malloc(sizeof(dm_options_t)), "no mem");
    
    argparser_t *pp = argparser_new(stderr, argc, argv);
    argparser_set_help(pp, PROG_HELP);
    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);
    
    o->symmetry = argparser_keyword_present(pp, "-symmetry");
    
    argparser_get_keyword(pp, "-scores");
    o->sc = dm_score_args_parse(pp);
    
    argparser_skip_parsed(pp);
    
    o->outName = argparser_get_next(pp);

    argparser_finish(pp);
    
    return o;
  }
